"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

# -*- coding: utf-8 -*-

import threading
import time


import numpy as np
from scipy.sparse import coo_array, bsr_matrix, coo_matrix, csr_matrix, csc_matrix, csr_array, csc_array

import qrisp.simulator.bi_array_helper as hlp


try:
    #sparse_dot_mkl seems to be only faster in situations, where the shape of the matrices,
    #results in better parallelization. For many quantum circuits, this was not the case
    #so this feature is disabled until further notice.
    raise Exception
    #Install command: conda install -c conda-forge sparse_dot_mkl
    from sparse_dot_mkl import dot_product_mkl
    def sparse_matrix_mult(a,b):
        
        if not isinstance(a, (np.ndarray, csr_array, csc_array)):
            a = a.tocsr()
        if not isinstance(b, (np.ndarray, csc_array, csr_array)):
            b = b.tocsc()
        
        return dot_product_mkl(a, b)
except:
    # print("Failed to import mkl sparse matrix multiplication. Install: with conda install -c conda-forge sparse_dot_mkl. Using scipy algorithm.")
    sparse_matrix_mult = lambda a,b : (a @ b)
    
#A quick helper function to evaluate the product of an iterable
def prod(iter):
    res = 1
    for i in range(len(iter)):
        res *= iter[i]
    return res


class BiArray:
    
    def v_dot(self, other):
        
        original_shape_self = self.shape
        orignal_shape_other = other.shape
        
        self.reshape((self.size,))
        other.reshape((other.size,))
        
        res = tensordot(self, other, (0,0))
        
        self.reshape(original_shape_self)
        other.reshape(orignal_shape_other)
        
        return res

#The classes inside this file aim describe so called BiArrays.
#A BiArray is an nd-array, where all the dimensions is are a power of 2, ie. 2, 4, 8, or 1024...
#This property can be used to efficiently manipulate indices of sparse matrices such that
#they can be processed quickly enough to use sparse matrix multiplication libraries for tensor contraction.
#This functionality is summarized in the SparseBiArray class.
#The DenseBiArray class is an interface to numpy arrays, that also use the restriction 
#of the dimension being a power of two. For instance does the DenseBiArray class support
#constant time reshaping and axis swapping. The reshaping and axis swappings are "collected"
#and executed together once the array is required for contraction.

#Both classes automatically make use of potential parallelism in the code.
#This is achieved by returning arrays where the content is "under construction" by a certain thread.
#If the "under construction" array is required in a contraction, the finishing of the calculation process is awaited.

#The tensordot function is a numpy-esque interface for performing tensor contractions. This function also 
#Automatically converts between sparse and dense BiArrays once a specified sparsity threshold is reached.

#A third class, the DummyBiArray, is a class that provides the same methods as the previous two,
#however the contractions are not actually executed, but instead the amount of floating point operations
#is logged in a class attribute. This allows to quickly evaluate the performance of a contraction order
#without actually contracting



multithreading_threshold = np.inf

#This SparseBiArray class provides a numpy-esque rank n tensor interface, which uses sparse matrix multiplication
#for tensor contraction. The fact that only powers of two are allowed for the size, allow efficient
#index manipulations, that permit faster construction and processing of said sparse matrices compared
#to for instance the rank n tensor interface of the "sparse" package.

#The idea is to store multi dimensional sparse arrays as the coo-array of their raveled version
#ie. we ravel the array and store the indices of the non-zero entries together with their data
#So for instance the array 
# array([[0, 2, 0, 0],
#        [3, 0, 0, 0]])
#is stored as with the non-zero indices nz_indices = [1, 4] and data = [2,3].
#If we now want to swap the axes of the array in this representation, we can do bit-level manipulations
#on the nz_indices array:
#In binary we have
#nz_indices = ["001", "100"]
#An axis swap can now be performed by manipulating the binary strings (which is possible rather efficiently)
#In this case the manipulation consists of moving the last two digits of the bitstring to the front
#After the axis swap we have
#nz_indices_axis_swapped = ["010", "001"] = [2, 1]
#If we now unravel the array with updated shape, we get
# array([[0, 3],
#        [2, 0],
#        [0, 0],
#        [0, 0]])
#Which confirms that this bit-level operation indeed represented an axis-swap

class SparseBiArray(BiArray):
    
    #This attribute decides after how many contractions, there should be a check for duplicates and zero entries
    #int the sparse matrix data
    contraction_counter_threshold = np.inf
    
    
    #Constructor of the SparseBiArray class.
    #init_object can be 
    #- a numpy array
    #- a tuple of arrays where init_object[0] is an array of non-zero indices and init_object[1] the corresponding data
    #- a sparse array in coo format
    def __init__(self, init_object, shape = None, contraction_counter = 0):
        #This attribute tracks how many sparse matrix contractions, this object has been through
        self.contraction_counter = contraction_counter
        
        
        #Handle the case of an array 
        if isinstance(init_object, np.ndarray):
            array = init_object
            
            
            #Store shape
            self.shape = array.shape
            
            #Ravel init array
            raveled_array = array.ravel()
            
            # raveled_array = np.around(raveled_array, 6, out = raveled_array)

            #Find non-zero indices
            self.nz_indices = np.nonzero(raveled_array)[0]
            
            #Store data
            self.data = raveled_array[self.nz_indices]
            
            # self.nz_indices, self.data = hlp.elim_zeros(self.nz_indices, self.data)
            
        #Handle the case that the init_object is a tuple of nz_indices/data arrays
        elif isinstance(init_object, tuple):
            
            if shape == None:
                raise Exception("Tried to initialise SparseBiArray from sparse data without providing a shape")
            
            self.nz_indices = init_object[0].ravel()
            self.data = init_object[1]
            
            self.nz_indices, self.data = hlp.elim_zeros(self.nz_indices, self.data)
        
        #Handle the case that the init_object is a sparse matrix/array
        elif isinstance(init_object, (coo_array, coo_matrix)):
            
            if shape == None:
                raise Exception("Tried to initialise SparseBiArray from sparse matrix without providing a shape")
            
            self.nz_indices = hlp.gen_flat_coords(init_object.col, init_object.row, init_object.shape)
            self.data = init_object.data
        
        elif isinstance(init_object, SparseBiArray):
            copy = init_object.copy()
            
            for var in copy.__dict__.keys():
            
                self.__dict__[var] = copy.__dict__[var]
                
            return
        else:
            raise Exception("Tried to initialize SparseBiArray with unknown type " + str(type(init_object)))
        if shape != None:
            self.shape = shape
        
        
        #Store the size
        self.size = prod(self.shape)
        
        if not ((self.size & (self.size-1) == 0) and self.size != 0):
            raise Exception("Tried to initialize SparseBiArray with size " + str(self.size) + " which is not a power of 2")
        
        #Store the sparsity
        self.sparsity = len(self.nz_indices)/self.size
        
        #This list keeps track of the axes permutations that have been applied
        #Once the SparseBiArray is needed for a contraction, all these swaps are applied within a single
        #function call and then the corresponding matrix is build from that
        self.index_bit_permutation = list(range(int(np.log2(self.size))))
        
        self.dtype = self.data.dtype
        
        #If the contraction counter is above the threshold, we sum the duplicates and eliminate the zeros
        if False:
        # if contraction_counter > self.contraction_counter_threshold:
            self.sum_duplicates()
            self.eliminate_zeros()
            self.contraction_counter = 0
    
    #Method for copying
    def copy(self):
        self.apply_swaps()
        return SparseBiArray((self.nz_indices.copy(), self.data.copy()), shape = tuple(self.shape))
    
    
    #This method checks if the sparse array has a thread object, which means that it's data
    #is under construction. If so, it waits until the thread is finished
    def catch_up(self):
        if hasattr(self, "thread"):
            start_time = time.time()
            self.thread.join()
            
            duration = time.time()-start_time
            if duration > 0.001:
                # print(duration)
                pass
            del self.thread
    
    
    #Reshaping method - note that no data is manipulated, ie. constant time
    def reshape(self, shape):
        
        if isinstance(shape, int):
            shape = (shape,)
        
        shape = tuple(shape)
        if prod(shape) != prod(self.shape):
            raise Exception("Tried to reshape with invalid shape")
        
        self.shape = shape
        
        return self
    
    
    #Method for swapping two axes
    def swapaxes(self, i, j):
        
        if i == j:
            return
        if i > j:
            i, j = j, i
        
        
        #In order to store the swap in the index_bit_permutation attribute,
        #we need to swap the corresponding digit ranges inside this list
        
        #For this we first determine the digit range of the index bitstrings
        #For instance imagine we have the index 215 and the shape (4,8,8)
        #and we want to swap the axes i = 0 and j = 2.
        #In binary, 215 is 11010111.
        #We first determine the layout of the indices inside this bitstring
        bit_shape = np.log2(self.shape).astype(np.int64)
        
        #In our case the bit shape is (2,3,3)
        
        i_interval_start = np.sum(bit_shape[:i])
        i_interval_end = i_interval_start + bit_shape[i]

        #In our example i = 0, the i_interval starts at the first (most significant) digit and ends after the third
        #ie |11|010111

        j_interval_start = np.sum(bit_shape[:j])
        j_interval_end = j_interval_start + bit_shape[j]
        
        #The j interval now starts at 6-th digit and ends at the last digit
        #ie. |11|010|111|
        
        #these two intervals have to be swapped.
        #However in order to makes this method constant in time, the actual swapping
        #of the indices is not performed here - rather we log the permutation of the indices
        
        #After creating of our SparseBiArray the attribute .index_bit_permutation simply holds
        #the numbers from 0 to log2(size)-1, ie. in the case of our example
        #[0,1,2,3,4,5,6,7]
        #In order to log the axis swap, we perform this permutation on the .index_bit_permutation attribute
        #ie. after the application of this method, we want it to be
        #[5,6,7,2,3,4,0,1]
        #ie. the interval of the first two and the last three swapped
        
        
        #In order to do this we split the index_bit_permutation list into 5 parts
        #1. the interval before the i-interval
        #2. the i-interval
        #3. the interval between i and j
        #4. the j interval
        #5. the interval after j
        
        #in the case of our example, we would have
        #1. = []
        #2. = [0,1]
        #3. = [2,3,4]
        #4. = [5,6,7]
        #5. = []
        
        #We than acquire the permuted list by concatenating these list in the correct order, ie.
        # 1. + 4. + 3. + 2. + 5.
        
        
        #Determine the first interval
        init_interval = self.index_bit_permutation[:i_interval_start]
        
        #Determine the i-interval
        i_interval = self.index_bit_permutation[i_interval_start:i_interval_end]
        
        #Determine the middle-interval
        middle_interval = self.index_bit_permutation[i_interval_end:j_interval_start]
        
        #Determine the j-interval
        j_interval = self.index_bit_permutation[j_interval_start:j_interval_end]
        
        #Determine the final interval
        final_interval = self.index_bit_permutation[j_interval_end:]
        
        #Concatenate intervals
        self.index_bit_permutation = init_interval + j_interval + middle_interval + i_interval + final_interval
        
        
        #Create the tuple for the new shape
        new_shape = list(self.shape)
                
        new_shape[i], new_shape[j] = new_shape[j], new_shape[i]
        
        self.shape = tuple(new_shape)
        
        return self
    
    #This method applies the index permutation generated too
    def apply_swaps(self):
        
        #Wait for any contraction processes to finish
        self.catch_up()
        
        #Check if any permutations have to be performed
        if self.index_bit_permutation == list(range(int(np.log2(self.size)))):
            return
        


        
        #Apply permutations
        #The bit level manipulations are done in the helper file
        self.nz_indices = hlp.permute_axes(self.nz_indices, np.array(self.index_bit_permutation)).ravel()
        
        #Reset the index_bit_permutation attribute, to 
        #indicate that no permutation has to be executed now
        self.index_bit_permutation = list(range(int(np.log2(self.size))))
    
    #This method sums duplicates ie. if an index appears twice in nz_indices,
    #the corresponding data entries are summed
    def sum_duplicates(self):
        self.nz_indices, self.data = hlp.sum_duplicates(self.nz_indices, self.data)
    
    #If after a contraction a new zero appeared, this entry does not need to tracked anymore
    #This method removes the zeros
    def elim_zeros(self):
        self.nz_indices, self.data = hlp.elim_zeros(self.nz_indices, self.data)
    
    
    #Builds up the sparse matrix in coo format
    #The coo format saves an array of row/column indices and the corresponding entries
    #Shape is the shape the sparse matrix should have
    def build_sr_matrix(self, shape, transpose = False):
        

        if prod(shape) != self.size:
            raise Exception("Tried to build sparse matrix with invalid shape")
        
        #Perform potential swaps
        self.apply_swaps()
        
        #Retrieve coordinates
        row, col = hlp.get_coordinates(self.nz_indices, shape)
        
        #Create sparse matrix
        if not transpose:
            res = coo_array((self.data, (row, col)), shape = shape)
        else:
            res = coo_array((self.data, (col, row)), shape = shape[::-1])
        
        
        #This attribute seems to be False which triggers a sorting algorithm,
        #which is in many cases among the most resource costly operations during simulation.
        #Setting this to True had no influence on the results tested so far.
        res.has_canonical_format = True
        
        return res
    
    #This method allows moving an index to the first position without
    #changing the order of the remaining indices
    def move_index_to_front(self, index):
        
        #if the index is already at the front, we don't need to do anything
        if index == 0:
            return
        
        #The logic for creating the ne self.index_bit_permutation attribute is similar as in swapaxes
        
        bit_shape = [int(np.log2(s)) for s in self.shape]
        
        new_shape = list(self.shape)
        new_shape.insert(0, new_shape.pop(index))
        
        self.shape = tuple(new_shape)
        
        index_range_start = sum(bit_shape[:index])
        index_range_end = index_range_start + bit_shape[index]
        
        init_perm = self.index_bit_permutation[:index_range_start]
        index_perm = self.index_bit_permutation[index_range_start:index_range_end]
        final_perm = self.index_bit_permutation[index_range_end:]
        self.index_bit_permutation = index_perm + init_perm + final_perm
        
    
    #This function performs tensor contraction along the specified axes
    #Axes can be contracted if they have the same dimension
    #Imagine we have two tensors with shapes (2,4,64,8,16) and (8,32,4,2)
    #then we could for instance contract the following axes
    #0 with 3
    #1 with 2
    #3 with 0
    #From this we will then get a tensor, which has the shape of the
    #first concatenated with the second but with the contraction indices removed
    #ie. (64,16,32)
    #The idea of this function is to move all the contraction indices to the front
    #reshape to matrices, transpose the first matrix, and then call some matrix 
    #multiplication algorithm. In the case of the example the shapes after 
    #moving to the front look like
    #(2,4,8,64,16) and (2,4,8,32)
    #After reshaping to matrices the shapes look like 
    #(2*4*8, 64*16) and (2*4*8, 32)
    #Transpose the first matrix
    #(64*16, 2*4*8) and (2*4*8, 32)
    #Call matrix multiplication to get a matrix of shape
    #(64*16,32)
    #Finally, reshape the result
    #(64,16,32)
    def contract(self, other, axes_self, axes_other):
        
        #Check if the contraction size of both tensors is allowed
        contraction_size = prod([self.shape[i] for i in axes_self])
        contraction_size_other= prod([other.shape[i] for i in axes_other])
        
        if contraction_size != contraction_size_other:
            raise Exception("Tried to contract tensor with differently sized contraction indices")
        
        #Copy axes list in order to prevent modification
        axes_self = list(axes_self)
        axes_other = list(axes_other)
        
        #Move axes to the front
        for i in range(len(axes_self)):
            self.move_index_to_front(axes_self[i])
            
            #Note that we need to increase the index of the axes
            #greater than the axes we just moved, because
            #the insertion at the front moved the all the indices
            #before axes_self[i]
            for j in range(i + 1, len(axes_self)):
                if axes_self[i] > axes_self[j]:
                    axes_self[j] += 1 
        
        #Perform the same for the other tensor
        for i in range(len(axes_other)):
            other.move_index_to_front(axes_other[i])
            
            for j in range(i + 1, len(axes_other)):
                if axes_other[i] > axes_other[j]:
                    axes_other[j] += 1
    
        
        #Determine new shape (now that the contraction axes are at the front, we can simply slice the shapes)
        res_shape = list(self.shape[len(axes_self):]) + list(other.shape[len(axes_other):])
        
        #In order to feature multiprocessing, we create an "empty" result array, which will be returned
        #after the calculation has been started. If the content is needed before the calculation finished,
        #the catch_up method will make sure the calculation finishes before further processing 
        res = SparseBiArray((np.zeros(1, dtype = np.int64), np.zeros(1, dtype = self.data.dtype)), shape = res_shape, contraction_counter = self.contraction_counter + other.contraction_counter)
        
        #Define multi processing wrapper
        def mp_wrapper():
            
            #Build up sparse matrices
            sr_matrix_self = self.build_sr_matrix((contraction_size, self.size//contraction_size), transpose = True)
            sr_matrix_other = other.build_sr_matrix((contraction_size, other.size//contraction_size))
            
            #Perform sparse matrix multiplication
            
            res_sr_matrix = sparse_matrix_mult(sr_matrix_self, sr_matrix_other).tocoo()
            
            
            #Acquire flattened coordinates from the helper function   
            res.nz_indices = hlp.gen_flat_coords(res_sr_matrix.col, res_sr_matrix.row, res_sr_matrix.shape)
            
            #Set the data
            res.data = res_sr_matrix.data

            # res.nz_indices, res.data = hlp.elim_zeros(res.nz_indices, res.data)

            
            #Set the sparsity
            res.sparsity = len(res.nz_indices)/res.size
            
            #Perform consolidation operations
            if False:
            # if res.contraction_counter > self.contraction_counter_threshold:
                res.sum_duplicates()
                res.eliminate_zeros()
                res.contraction_counter = 0
        
        #Start the thead
        if res.size > multithreading_threshold:
            res.thread = threading.Thread(target = mp_wrapper)
            res.thread.start()
        else:
            mp_wrapper()
        
        return res 

    #This method returns self as a numpy array        
    def to_array(self):
        #Apply all swapping operations
        self.apply_swaps()
        
        #Construct the flat array
        res = hlp.construct_flat_array(self.nz_indices, self.data, self.size)
        
        #Reshape
        return res.reshape(self.shape)

    
    
    def ravel(self):
        return self.reshape(self.size)
    
    def __repr__(self):
        return str(self.to_array())
    
    
    #This function allows to split the array into it's lower and upper half
    #ie. if this SparseBiArray represents the array [0,1,2,3,4,5,6,7]
    #we return [0,1,2,3] and [4,5,6,7] as SparseBiArrays
    #This feature is important when measuring QuantumState as the norms of the
    #lower and upper half of the array determine the probability of measuring a 1 or 0
    def split(self):
        
        #Check if self is in a suitable shape (multi dimensional arrays can not be splitted as intended)
        if len(self.shape) != 1:
            raise Exception("Tried to split multi dimensional sparse array")
            
        #Applya swaps
        self.apply_swaps()    
        
        
        #Find the boolean mask of indizes that are less than half of the size
        mask = self.nz_indices < self.size//2
        
        #Acquire the lower half nz_indices
        lower_half_nz_indices = self.nz_indices[mask].copy()
        #Acquire the lower half data
        lower_half_data = self.data[mask].copy()
        
        #Create lower half result
        lower_half = SparseBiArray((lower_half_nz_indices, lower_half_data), shape = (self.shape[0]//2,))
        
        #Perform the same process for the upper half but with inverted mask
        mask = np.invert(mask)
        upper_half_nz_indices = self.nz_indices[mask] - self.size//2
        upper_half_data = self.data[mask].copy()
        
        upper_half = SparseBiArray((upper_half_nz_indices, upper_half_data), shape = (self.shape[0]//2,))
        
        #Return result
        return lower_half, upper_half
    
    def multi_split(self, indices):
        
        sprs = self.build_sr_matrix([2**(len(indices)), self.size//2**len(indices)]).tocsr()
        
        col_indices = sprs.indices
        row_ptr = sprs.indptr
        data = sprs.data
        
        new_bi_arrays = []
        for r in range(len(row_ptr)-1):
            if row_ptr[r] == row_ptr[r+1]:
                new_bi_array = None
            else:
                new_bi_array = SparseBiArray((col_indices[row_ptr[r]:row_ptr[r+1]].astype(np.int64), data[row_ptr[r]:row_ptr[r+1]]), shape = (self.size//2**len(indices),))
            
            new_bi_arrays.append(new_bi_array)
            
            
        
        return new_bi_arrays
    
    #Calculate the squared norm, ie. the sesquilinear scalar product of self with itself
    def squared_norm(self):
        return np.abs(np.vdot(self.data, self.data))
    
    
    #Return self as a DenseBiArray
    def to_dense(self):
        
        res = DenseBiArray(np.zeros(1), sparsity = self.sparsity, shape = self.shape)
        #We perform the same technique to enable multithreading as in contract,
        #ie. return an unfinished result that is waited on if required
        def mp_wrapper():
            res.data = self.to_array()
        
        
        #Start the thead
        if self.size > multithreading_threshold:
            res.thread = threading.Thread(target = mp_wrapper)
            res.thread.start()
        else:
            mp_wrapper()
        
        return res
    
    def to_sparse(self):
        return self


#This class serves mainly as an interface to numpy arrays using the same methods as the SparseBiArray class
#this way the algorithm using these classes doesn't need to care about wether it's treating
#a sparse array or a dense array.
#Furthermore some of the multithreading techniques from the SparseBiArray class are deployed
class DenseBiArray(BiArray):
    
    #The constructor can only be called with numpy arrays
    def __init__(self, array, sparsity = None, shape = None):
        
        if not isinstance(array, np.ndarray):
            raise Exception("Tried to initialize DenseBiArray with type " + str(type(array)))
        
        #Save data
        self.data = array
        
        #Set shape
        if shape:
            self.shape = shape
        else:
            self.shape = array.shape
            
        #Determine size
        self.size = prod(self.shape)
        
        #If not given, estimate sparsity. In many situations the sparsity
        #of a contraction result can be estimated (and then given to the constructor), 
        #so it is helpfull to track it
        if sparsity == None:
            self.sparsity = np.count_nonzero(array)/array.size
        else:
            # self.sparsity = np.count_nonzero(array)/array.size
            self.sparsity = sparsity
        
        #This attribute serves a similar purpose as in the SparseBiArray class
        self.index_bit_permutation = list(range(int(np.log2(self.size))))
        
        self.dtype = self.data.dtype
    
    #Copy method
    def copy(self):
        self.apply_swaps()
        return DenseBiArray(self.data.copy(), sparsity = self.sparsity)
    
    #Catch up method (same purpose as in SparseBiArray)
    def catch_up(self):
        if hasattr(self, "thread"):
            start_time = time.time()
            self.thread.join()
            
            duration = time.time()-start_time
            if duration > 0.001:
                # print("Waited ", duration)
                pass
            del self.thread
        
    #The shape of the DenseBiArray is tracked only as a tuple.
    #Once the data is required for a contraction, the actual data will be reshaped
    def reshape(self, shape):
        
        if isinstance(shape, int):
            shape = (shape,)
        
        if prod(shape) != prod(self.shape):
            raise Exception("Tried to reshape with invalid shape")
        
        self.shape = tuple(shape)
        return self
    
    #This method works similarly as it's equivalent in SparseBiArray
    def swapaxes(self, i, j):
        
        
        if i == j:
            return
        if i > j:
            i, j = j, i
        
        
        bit_shape = np.log2(self.shape).astype(np.int64)
        
        i_interval_start = sum(bit_shape[:i])
        i_interval_end = i_interval_start + bit_shape[i]

        j_interval_start = sum(bit_shape[:j])
        j_interval_end = j_interval_start + bit_shape[j]
        
        init_interval = self.index_bit_permutation[:i_interval_start]
        i_interval = self.index_bit_permutation[i_interval_start:i_interval_end]
        
        middle_interval = self.index_bit_permutation[i_interval_end:j_interval_start]
        
        j_interval = self.index_bit_permutation[j_interval_start:j_interval_end]
        
        final_interval = self.index_bit_permutation[j_interval_end:]
        
        self.index_bit_permutation = init_interval + j_interval + middle_interval + i_interval + final_interval
        
        new_shape = list(self.shape)
                
        new_shape[i], new_shape[j] = new_shape[j], new_shape[i]
        
        self.shape = tuple(new_shape)
        
        return self
    
    #This method works similarly as it's equivalent in SparseBiArray
    def move_index_to_front(self, index):
        
        
        if index == 0:
            return
        
        bit_shape = [int(np.log2(s)) for s in self.shape]
        
        new_shape = list(self.shape)
        new_shape.insert(0, new_shape.pop(index))
        
        self.shape = tuple(new_shape)
        
        index_range_start = sum(bit_shape[:index])
        index_range_end = index_range_start + bit_shape[index]
        
        init_perm = self.index_bit_permutation[:index_range_start]
        index_perm = self.index_bit_permutation[index_range_start:index_range_end]
        final_perm = self.index_bit_permutation[index_range_end:]
        self.index_bit_permutation = index_perm + init_perm + final_perm
        
    #This method works similarly as it's equivalent in SparseBiArray
    def apply_swaps(self):
        
        #Wait for any computation processes to finish
        self.catch_up()
        
        #Check if axes need to be swapped
        if self.index_bit_permutation == list(range(int(np.log2(self.size)))):
            
            #Check if data needs to be reshaped
            if self.data.shape != self.shape:
                self.data = self.data.reshape(self.shape)
        
            return
        
        #The plan is now to reshape into "tensor format" ie. log2(self.size) axes of dimension
        #two and permute the axes
        
        #Numpy however does not allow arrays with more than 32 axes, so it would not be possible
        #to manipulate arrays describing more than 32 qubits.
        #In order to reduce the number of axes that need to be permuted,
        #we identify intervals of indices that do not need permutation on them
        #and group them into one axis
        
        
        
        
        #Get the permutation of the axes
        bi_index_perm = [self.index_bit_permutation.index(i) for i in range(int(np.log2(self.size)))]
        
        
        
        if False:
            
            
            # result = np.empty(self.data.size, dtype = self.data.dtype)

            index_perm = hlp.invert_permutation(np.array(bi_index_perm))[::-1]
            index_perm = np.abs(len(index_perm)-1-index_perm)
            
            
            #flattened index array
            f_index_array = np.empty(2**len(index_perm), dtype = np.int64)
            
            f_index_array[0] = 0
            # result[0] = self.data.ravel()[0]
            
            
            #Unfortunately, this is not faster (maybe if implemented in C?)
            self.data = hlp.bi_array_moveaxis(self.data.ravel(), index_perm, f_index_array)
            # temp = hlp.bi_array_moveaxis(self.data.ravel(), index_perm, f_index_array)
            
            # permutated_indices = hlp.bi_array_moveaxis_jitted(index_perm, f_index_array)
            # self.data = self.data.ravel()[temp.ravel()]
            
            self.data.shape = self.shape
        
        else:
            perm = list(bi_index_perm)
            #Initialize the shape
            perm_shape = len(self.index_bit_permutation)*[2]
            
            
            i = 1
            while i < len(perm):
                #If this condition is met, both indices form an interval,
                #that does not need further permutation
                if perm[i]-1 == perm[i-1]:
                    
                    #Remove the index from the perm and perm_shape list
                    perm.pop(i)
                    perm_shape.pop(i)
                    
                    #Update the shape
                    perm_shape[i-1] *= 2
                    
                    #Update the permutation indices
                    for j in range(len(perm)):
                        if perm[j] > perm[i-1]:
                            perm[j] -= 1
                else:
                    i += 1
            
                
            #Reshape data to the required shape
            self.data = self.data.reshape(perm_shape)
            
            #Move the axes on the data
            self.data = np.moveaxis(self.data, list(range(len(perm))), perm)
        
            #Reshape the data
            self.data = self.data.reshape(self.shape)
        
        #Set the index_bit_permutation to identity
        self.index_bit_permutation = list(range(int(np.log2(self.size))))
    
    #Raveling method
    def ravel(self):
        self.reshape(self.size)
        self.apply_swaps()
        return self
        
    #This function works similar as it's equivalent in SparseBiArray
    #ie. identify contraction axes, move to the front, transpose one matrix
    #and then apply matrix multiplication
    def contract(self, other, axes_self, axes_other):
        
        
        #Check if the contraction size of both tensors is allowed
        contraction_size = prod([self.shape[i] for i in axes_self])
        contraction_size_other= prod([other.shape[i] for i in axes_other])
        
        if contraction_size != contraction_size_other:
            raise Exception("Tried to contract tensor with differently sized contraction indices")
        
        #Copy axes list in order to prevent modification
        axes_self = list(axes_self)
        axes_other = list(axes_other)
        
        #Move axes to the front
        for i in range(len(axes_self)):
            self.move_index_to_front(axes_self[i])
            
            #Note that we need to increase the index of the axes
            #greater than the axes we just moved, because
            #the insertion at the front moved the all the indices
            #before axes_self[i]
            for j in range(i + 1, len(axes_self)):
                if axes_self[i] > axes_self[j]:
                    axes_self[j] += 1 
        
        #Perform the same for the other tensor
        for i in range(len(axes_other)):
            other.move_index_to_front(axes_other[i])
            
            for j in range(i + 1, len(axes_other)):
                if axes_other[i] > axes_other[j]:
                    axes_other[j] += 1
    
        
        #Determine new shape (now that the contraction axes are at the front, we can simply slice the shapes)
        res_shape = list(self.shape[len(axes_self):]) + list(other.shape[len(axes_other):])
        
        
        #The sparsity can be estimated using statistical methods:
        #In an regular tensor contraction, any entry of the resulting tensor,
        #is the sum of n = contraction_size products of entries of the two source tensors
        #Each of these products is non-zero with probability p_nz = self.sparsity*other.sparsity
        
        # p_nz = self.sparsity*other.sparsity
        
        #The probabilty for the product being zero is therefore p_z = 1 - self.sparsity*other.sparsity
        
        # p_z = 1 - self.sparsity*other.sparsity
        
        #therefore we perform n = contraction_size random experiments with probability p_z
        #The result entry is zero if all products are zero, so
        
        # p_res_z = p_z**contraction_size
        
        #To get the sparsity, we finally need the probability of the result entry being non-zero
        
        # res_sparsity = 1-p_res_z
        
        #We tested this formula and it works extremely well.
        
        res_sparsity = 1-(1-self.sparsity*other.sparsity)**contraction_size
        
        #Prepare "unfinished" result
        res = DenseBiArray(np.array([self.data.ravel()[0]*other.data.ravel()[0]]), sparsity = res_sparsity, shape = res_shape)
        
        #Prepare multithreading wrapper
        #Performs the same logic as it's SparseBiArray equivalent
        def mp_wrapper():
            
            
            original_shape_self = self.shape
            original_shape_other = other.shape
            
            self.reshape([contraction_size, self.size//contraction_size])
            other.reshape([contraction_size, other.size//contraction_size])
            
            self.swapaxes(0,1)
            
            self.apply_swaps()
            other.apply_swaps()
            
            res.data = np.matmul(self.data, other.data)
            
            self.swapaxes(0,1)
            self.reshape(original_shape_self)
            other.reshape(original_shape_other)
            
            
        if res.size > multithreading_threshold:
            #Start the wrapper    
            res.thread = threading.Thread(target = mp_wrapper)
            res.thread.start()
        else:
            mp_wrapper()
        #Return the result
        return res
    
    #This method works similarly as it's equivalent in SparseBiArray
    def split(self):
        
        if len(self.shape) != 1:
            raise Exception("Tried to split multi rank BiArray")
        self.apply_swaps()    
        
        return DenseBiArray(self.data[:self.size//2].copy(), self.sparsity), DenseBiArray(self.data[self.size//2:].copy(), sparsity = self.sparsity)
    
        
    def multi_split(self, indices):
        
        original_shape = tuple(self.shape)
        self.reshape([2**(len(indices)), self.size//2**len(indices)])
        
        np_array = self.to_array()
        
        new_bi_arrays = []
        
        for r in range(2**len(indices)):
            new_bi_arrays.append(DenseBiArray(np_array[r,:]))
        
        self.reshape(original_shape)
        
        return new_bi_arrays

    #This method works similarly as it's equivalent in SparseBiArray
    def squared_norm(self):
        self.apply_swaps()
        return np.abs(np.vdot(self.data, self.data))
    
    #This method works similarly as it's equivalent in SparseBiArray
    def to_array(self):
        self.apply_swaps()
        return self.data
    
    
    #This method works similarly as it's equivalent in SparseBiArray    
    def to_dense(self):
        return self
    
    #Generates the SparseBiArray version of self using multithreading
    def to_sparse(self):
        
        res = SparseBiArray((np.zeros(1, dtype = np.int64), np.zeros(1)), shape = self.shape)

        #Prepare the wrapper       
        def mp_wrapper():
            self.apply_swaps()
            
            #Ravel init array
            raveled_array = self.data.ravel()
            
            
            #Find non-zero indices
            res.nz_indices = np.nonzero(raveled_array)[0]

            #Store data
            res.data = raveled_array[res.nz_indices]
            
            res.nz_indices, res.data = hlp.elim_zeros(res.nz_indices, res.data)
            
            res.sparsity = len(res.nz_indices)/res.size
        
        #Start the
        if res.size > multithreading_threshold:
            res.thread = threading.Thread(target = mp_wrapper)
            res.thread.start()
        else:
            mp_wrapper()        
        return res
    
    def __repr__(self):
        return str(self.to_array())

#This class is used to estimate the complexity of a contraction order
#The contract method does not actually execute a contraction but only logs
#the estimated amount of floating point operations
class DummyBiArray(BiArray):
    
    current_memory = np.zeros(1)
    max_memory = np.zeros(1)
    flops = np.zeros(1)
    
    def __init__(self, data = None, shape = None, sparsity = 1):
        
        self.sparsity = sparsity
        if not isinstance(data, type(None)):
            self.shape = data.shape
            self.sparsity = np.count_nonzero(data)/data.size
        if not isinstance(shape, type(None)):
            self.shape = tuple(shape)
        
        
        self.size = prod(self.shape)
        
    def reshape(self, shape):
        self.shape = shape
        
    def swapaxes(self, axis_0, axis_1):
        
        self.shape = list(self.shape)
        self.shape[axis_0], self.shape[axis_1] = self.shape[axis_1], self.shape[axis_0]
        self.shape = tuple(self.shape)
    
    def contract(self, other, axes_self, axes_other):
        
        contraction_size = prod([self.shape[i] for i in axes_self])
        
        temp_shape_self = list(self.shape)
        temp_shape_other = list(other.shape)
        
        delete_multiple_element(temp_shape_self, axes_self)
        delete_multiple_element(temp_shape_other, axes_other)
        
        res_sparsity =  1-(1-self.sparsity*other.sparsity)**contraction_size
        res_shape = temp_shape_self + temp_shape_other
        try:
            self.flops += prod(res_shape)*contraction_size
        except TypeError:
            self.flops += np.inf
            
        res = DummyBiArray(shape = res_shape, sparsity = res_sparsity)
        self.current_memory[0] += res.size - self.size - other.size
        
        if self.current_memory > self.max_memory:
            self.max_memory[0] = self.current_memory[0]
        
        return res
        
    def copy(self):
        return DummyBiArray(shape = self.shape)
    @classmethod
    def reset_stat_counter(self):
        
            
        self.current_memory = np.zeros(1)
        self.max_memory = np.zeros(1)
        self.flops = np.zeros(1)


#This function emulates the behavior of np.tensordot
def tensordot(a,b, axes, contract_sparsity_threshold = 0.01):
    #Turn axes object into a list
    axes = list(axes)
    if isinstance(axes[0], int):
        axes[0] = [axes[0]]
    if isinstance(axes[1], int):
        axes[1] = [axes[1]]
    
    #Treat the case of a dummy contraction    
    if isinstance(a, DummyBiArray) and isinstance(b, DummyBiArray):
        return a.contract(b, axes[0], axes[1])
    
    #Treat the case of a numpy contraction
    if isinstance(a, np.ndarray) and isinstance(b, np.ndarray):
        return np.tensordot(a,b, axes)
    
    if a.data.dtype == np.dtype("O") or b.data.dtype == np.dtype("O"):
        a = a.to_dense()
        b = b.to_dense()
        
        return a.contract(b, axes[0], axes[1])
    
    if len(axes[0]) != 0:
        
        contraction_size = prod([a.shape[i] for i in axes[0]])
        
        #Estimate resulting sparsity (for more information on this estimate check the contract method of DenseBiArray)
        res_sparsity = 1-(1-a.sparsity*b.sparsity)**contraction_size
        
        res_size = a.size*b.size/contraction_size
    else:
        res_sparsity = a.sparsity*b.sparsity
        
        res_size = a.size*b.size
    
    
    
    if (res_sparsity > contract_sparsity_threshold or a.size*b.size < 2**12) and not res_size > 2**32:
        
    #     if ((isinstance(a, SparseBiArray) and isinstance(b, DenseBiArray)) or (isinstance(b, SparseBiArray) and isinstance(a, DenseBiArray))):
            
    #         return mixed_contractor(a, b, axes)
    
            a = a.to_dense()
            b = b.to_dense()
        
    else:
            
            a = a.to_sparse()
            b = b.to_sparse()
        
        
    #Treat the case of a "contractionless" tensor product    
    if len(axes[0]) == 0:
        
        #We utilize the multithreading pattern presented in the previous contraction methods
        if isinstance(a, SparseBiArray) and isinstance(b, SparseBiArray):
            #Prepare result arrays            
            res = SparseBiArray((np.zeros(1, dtype = np.int64), np.ones(1)), shape = list(a.shape) + list(b.shape))
            
            #Prepare wrapper
            def mp_wrapper():
                
                #Apply any potential swaps on the source arrays
                a.apply_swaps()
                b.apply_swaps()
                
                #Imagine we have two arrays with flattened nz_indices [1,2,3], [2,3,6] and sizes 16, 8  respectively
                #the nz_indices of the result can be ordered in a matrix
                #[1,2,3] + 2*16
                #[1,2,3] + 3*16
                #[1,2,3] + 6*16
                #(where the plus is executed on all entries)
                
                #This idea is encapsulated in the following command
                nz_indices = np.tensordot(b.size*a.nz_indices, np.ones(len(b.data), dtype = np.int64), ((),())) + np.tensordot(np.ones(len(a.data), dtype = np.int64), b.nz_indices, ((),()))
                
                #The corresponding data can be calculated as
                data = np.tensordot(a.data.ravel(), b.data.ravel(), ((),())).ravel()
                
                #Set nz_indices and data
                res.nz_indices = nz_indices.ravel()
                
                res.data = data
            
            if res.size > multithreading_threshold:
                res.thread = threading.Thread(target = mp_wrapper)
                res.thread.start()
            else:
                mp_wrapper()
                
            return res
            
        else:
            #For the DenseBiArray, we can simply use the numpy tensordot
            a.catch_up()
            b.catch_up()
            
            res = DenseBiArray(np.array([a.data.ravel()[0]*b.data.ravel()[0]]), sparsity = a.sparsity*b.sparsity, shape = list(a.shape) + list(b.shape))
            
            def mp_wrapper():
                a.apply_swaps()
                b.apply_swaps()
                
                res.data = np.tensordot(a.to_array(), b.to_array(), axes = ((),()))
            
            if res.size > multithreading_threshold:
                res.thread = threading.Thread(target = mp_wrapper)
                res.thread.start()
            else:
                mp_wrapper()
            
            return res
    
    return a.contract(b, axes[0], axes[1])


def mixed_contractor(a, b, axes):
    
    if not ((isinstance(a, SparseBiArray) and isinstance(b, DenseBiArray)) or (isinstance(b, SparseBiArray) and isinstance(a, DenseBiArray))):
        raise Exception("Tried to call mixed contraction on non-mixed input")
        
    #Copy axes list in order to prevent modification
    axes_a = list(axes[0])
    axes_b = list(axes[1])
        
        
    #Check if the contraction size of both tensors is allowed
    contraction_size = prod([a.shape[i] for i in axes_a])
    contraction_size_b= prod([b.shape[i] for i in axes_b])
    
    if contraction_size != contraction_size_b:
        raise Exception("Tried to contract tensor with differently sized contraction indices")
    
    
    #Move axes to the front
    for i in range(len(axes_a)):
        a.move_index_to_front(axes_a[i])
        
        #Note that we need to increase the index of the axes
        #greater than the axes we just moved, because
        #the insertion at the front moved the all the indices
        #before axes_a[i]
        for j in range(i + 1, len(axes_a)):
            if axes_a[i] > axes_a[j]:
                axes_a[j] += 1 
    
    #Perform the same for the b tensor
    for i in range(len(axes_b)):
        b.move_index_to_front(axes_b[i])
        
        for j in range(i + 1, len(axes_b)):
            if axes_b[i] > axes_b[j]:
                axes_b[j] += 1

    
    #Determine new shape (now that the contraction axes are at the front, we can simply slice the shapes)
    res_shape = list(a.shape[len(axes_a):]) + list(b.shape[len(axes_b):])
    
    
    res_sparsity = 1-(1-a.sparsity*b.sparsity)**contraction_size
    
    #Prepare "unfinished" result
    res = DenseBiArray(np.empty(1, dtype = a.data.dtype), sparsity = res_sparsity, shape = res_shape)
    
    
    #Define multi processing wrapper
    def mp_wrapper():
        
        #Build up sparse matrices
        if isinstance(a, SparseBiArray):
            mult_matrix_a = a.build_sr_matrix((contraction_size, a.size//contraction_size), transpose = True)
            original_shape_b = b.shape
            b.reshape([contraction_size, b.size//contraction_size])
            b.apply_swaps()
            mult_matrix_b = b.data
            
        else:
            
            original_shape_a = a.shape
            a.reshape([contraction_size, a.size//contraction_size])
            a.swapaxes(0,1)
            a.apply_swaps()
            mult_matrix_a = a.data
            mult_matrix_b = b.build_sr_matrix((contraction_size, b.size//contraction_size))
        
        
        #Set the data
        res.data = sparse_matrix_mult(mult_matrix_a, mult_matrix_b)
    
    #Start the thead
    if res.size > multithreading_threshold:
        res.thread = threading.Thread(target = mp_wrapper)
        res.thread.start()
    else:
        mp_wrapper()
    
    return res

def delete_multiple_element(list_object, indices):
    indices = sorted(indices, reverse=True)
    for idx in indices:
        if idx < len(list_object):
            list_object.pop(idx)



"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""



#TO-DO implement all the concepts discussed in https://stackoverflow.com/questions/30081275/why-is-1000000000000000-in-range1000000000000001-so-fast-in-python-3?rq=1

from qrisp.environments.quantum_conditionals import ConditionEnvironment, quantum_condition
from qrisp.logic_synthesis.truth_tables import TruthTable
from qrisp import x, cx, control, perm_lock, perm_unlock
from qrisp.core import QuantumVariable

class QuantumIterator:
    
    def __init__(self, quantum_condition, init_index = None):
        
        self.quantum_condition = quantum_condition
        
        
    def __iter__(self):
    	
        self.quantum_condition.start_dumping()
        
        return self
    
    def __next__(self):
        
        raise Exception("__next__ method for abstract quantum iterator class not defined")
        
        
        
        
class qRange:
    
    def __init__(self, max_index_qf, create_index_qf = False):

        from qrisp.qtypes.quantum_float import QuantumFloat
        if not isinstance(max_index_qf, QuantumFloat):
            raise Exception("Can only create quantum iterators from quantum variables")

        
        if max_index_qf.exponent != 0:
            raise Exception("qRange can only be intialized with integer quantum indices")
        
        self.create_index_qf = create_index_qf
        self.max_index_qf = max_index_qf
        
        
    def __iter__(self):
        
        if self.create_index_qf:
            self.index_qf = self.max_index_qf.duplicate(qs = self.max_index_qf.qs, name = "index_qf")
            self.index_qf.init_from(self.max_index_qf)
                
        else:
            self.index_qf = self.max_index_qf
            
        if not self.index_qf.signed:
            self.index_qf.extend(1, position = -1)
        
        
        x(self.index_qf)
        
        self.quantum_condition_env = control(self.index_qf[-1])
        
        self.c_index = -1
        
        perm_lock(self.index_qf)
        
        return self
        
    def __next__(self):
        
        perm_unlock(self.index_qf)
        self.c_index += 1
        
        if self.c_index != 0:
            
            
            self.quantum_condition_env.__exit__(None, None, None)
            

            if self.c_index >= 2**(self.index_qf.size-1):
                    
                if self.create_index_qf:
                    
                    x(self.index_qf)
                    self.index_qf += self.c_index - 1
                    
                    if not self.max_index_qf.signed:
                        cx(self.max_index_qf, self.index_qf[:-1])
                    else:
                        cx(self.max_index_qf, self.index_qf)
                    self.index_qf.delete()
                else:
                    
                    x(self.index_qf)
                    self.index_qf += self.c_index - 1
                    
                    self.index_qf.reduce(self.index_qf[-1])
                
                perm_unlock(self.index_qf)
                raise StopIteration
        
            self.index_qf += 1
        # @quantum_condition
        # def neg_condition(sign_qubit):
        #     return sign_qubit
            
        # self.quantum_condition_env = neg_condition(self.index_qf[-1])
        
        
        self.quantum_condition_env.__enter__()
        perm_lock(self.index_qf)
        return int(self.c_index)
        


"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

# Created by ann81984 at 04.05.2022
import pytest
import time

from qrisp.circuit import multi_controlled_circuit, RYGate, RZGate, RXGate, transpile, HGate, XGate, PGate
from qrisp import QuantumSession, QuantumVariable, QuantumCircuit, transpile, x
from qiskit.circuit.library import HGate

# Check by printing circuit
def test_controlled_gates():
    
    qc = QuantumCircuit(3)

    simple_case = False

    if simple_case:
        qc.x(0)
    else:
        # qc.h(0)
        # qc.cx(0,1)
        qc.cp(2.4, 0, 1)


    # print(qs)

    controlled_qc = multi_controlled_circuit(qc, 1)
    print(controlled_qc)



    ###################
    # Check by measuring circuit
    qs_0 = QuantumSession()

    qv = QuantumVariable(2, qs_0)

    if simple_case:
        qs_0.x(0)
    else:
        qs_0.h(0)
        qs_0.cx(0, 1)
        qs_0.cp(2.4, 0, 1)

    operation = qs_0.to_op()

    controlled_operation = operation.control(1)

    control_qv = QuantumVariable(1)
    qv = QuantumVariable(operation.num_qubits)

    control_on = True
    if control_on:
        x(control_qv[0])

    control_qv.qs.append(controlled_operation, control_qv.reg + qv.reg)

    tmp = qv.get_measurement()
    print(tmp)

    assert 0.4 < tmp['11'] < 0.6 and 0.4 < tmp['00'] < 0.6
    assert not (0.8 < tmp['11'] < 1 and 0.7 < tmp['00'] < 0.2)
    assert isinstance(tmp, dict)
    with pytest.raises(KeyError) as excinfo:
        bool(tmp['23'] < 0.6)
    assert '23' in str(excinfo.value)



    ###################
    start_time = time.time()
    
    n = 5
    test_gate = XGate().control(n)
    # test_gate = XGate().control(n)

    qs = QuantumSession()

    qv_ctrl = QuantumVariable(n, qs)
    qv_target = QuantumVariable(1, qs)

    x(qv_ctrl)

    qs.append(test_gate, qs.qubits)

    # Can be sped up to super-qiskit performance by enabling use_gray code
    # in logic_synthesis/gray_synthesis.py (this costs cnot count however)
    print("Took " + str(time.time() - start_time) + " s for gate synthesis")


    print(qv_target.get_measurement())
    tmp2 = qv_target.get_measurement()
    assert isinstance(tmp2, dict)
    assert tmp2['1'] == 1.0


    qv_0 = QuantumVariable(2)
    qv_1 = QuantumVariable(1)


    qv_0[:] = "10"
    qv_0.qs.append(XGate().control(2, ctrl_state = "10"), qv_0.reg + qv_1.reg)

    assert qv_1.get_measurement() == {"1" : 1.0}
    
    qv_0 = QuantumVariable(4)
    qv_1 = QuantumVariable(1)
    
    qv_0[:] = "0110"
    qv_0.qs.append(XGate().control(2, ctrl_state = "10").control(2, ctrl_state = "01"), qv_0.reg + qv_1.reg)
    
    assert qv_1.get_measurement() == {"1" : 1.0}
    
    
    
    



"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

# Created by ann81984 at 06.05.2022
# import pytest


def test_qompiler():
    from qrisp import QuantumSession, QuantumVariable, x, cx

    qs = QuantumSession()

    qv_list = [QuantumVariable(1, qs = qs) for i in range(10)]

    for i in range(9):
        x(qv_list[i])
        x(qv_list[i])
        qv_list[i].delete()

    x(qv_list[-1])
    qc = qs.compile()
    assert len(qc.qubits) == 1



"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

class Qubit:
    """
    This class describes qubits. Qubits are created by supplying the identifier string.
    
    Attributes
    ----------
    identifier : str
        A string to identify the Qubit.
    
    Examples
    --------
    
    We create a Qubit and add it to a :ref:`QuantumCircuit`:
    
    >>> from qrisp import QuantumCircuit, Qubit
    >>> qb = Qubit("alphonse")    
    >>> qc = QuantumCircuit()
    >>> qc.add_qubit(qb)
    >>> qc.x(qb)
    >>> print(qc)
              ┌───┐
    alphonse: ┤ X ├
              └───┘
    

    """
    def __init__(self, identifier):
        self.identifier = identifier
        self.hash_value = id(self)
        self.lock = False
        self.perm_lock = False
        
    def __str__(self):
        return self.identifier
    def __repr__(self):
        return "Qubit(" + self.identifier + ")"
    def __hash__(self):
        return self.hash_value
    def __eq__(self, other):
        return self.hash_value == other.hash_value



"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

from qrisp.circuit import QuantumCircuit, transpile, QubitAlloc, QubitDealloc
from qrisp import QuantumSession, QuantumFloat, QuantumChar, QuantumBool, merge
from qrisp.quantum_network import QuantumNetworkClient
import numpy as np

class QuantumNetworkSession(QuantumSession):
    """
    This class allows to use Qrisps high-level programming interface to interact with QuantumNetworks. As an inheritor of the QuantumSession class, we can create QuantumVariables with it which can then be sent to other participators of the QuantumNetwork.

    Note that get_measurement calls are also realized by the single-shot simulator. For more information on this check the documentation of the run method of the QuantumNetworkClient class.

    Parameters
    ----------
    name : str
        A nickname such as "Alice" or "Bob", which is used to direct the communication of other users to you.
    socket_ip : str
        The IP of the QuantumNetworkServer to connect to.
    port : int
        The port of the QuantumNetworkServer to connect to. The default is 7070.

    Examples
    --------
    
    We create a QuantumNetworkServer and connect a QuantumNetworkSession to it.
    
    >>> from qrisp.quantum_network import QuantumNetworkServer, QuantumNetworkSession
    >>> example_server = QuantumNetworkServer(socket_ip = "127.0.0.1", background = True)
    >>> example_server.start()
    >>> alice_session = QuantumNetworkSession(socket_ip = "127.0.0.1", port = 7070, name = "alice")
    
    Now we can create QuantumVariables as we are used to
    
    >>> from qrisp import QuantumFloat
    >>> qf = QuantumFloat(4, 5, qs = alice_session)

    
    """
    
    def __init__(self, name, socket_ip, port = None):
        
        qn_client = QuantumNetworkClient(name = name, socket_ip = socket_ip, port = port)
        self.inbox = []
    
        super().__init__(backend = qn_client)
        
    
    
    def send_qv(self, qv, recipient):
        """
        Sends a :ref:`QuantumVariable` to another participant of the network.

        Parameters
        ----------
        qv : QuantumVariable
            The QuantumVariable to send.
        recipient : str
            The name of the recipient.

        Examples
        --------
        
        We create a QuantumNetworkServer and connect a QuantumNetworkSession to it.
        
        >>> from qrisp.quantum_network import QuantumNetworkServer, QuantumNetworkSession
        >>> example_server = QuantumNetworkServer(socket_ip = "127.0.0.1", background = True)
        >>> example_server.start()
        >>> alice_session = QuantumNetworkSession(socket_ip = "127.0.0.1", port = 7070, name = "alice")
        >>> bob_session = QuantumNetworkSession(socket_ip = "127.0.0.1", port = 7070, name = "bob")
        
        We create a :ref:`QuantumFloat` and perform some arithmetic
        
        >>> from qrisp import QuantumFloat
        >>> qf = QuantumFloat(4, qs = alice_session)
        >>> qf += 3
        >>> alice_session.send_qv(qf, "bob")
        

        """
        
        if not qv.qs == self:
            merge(self, qv.qs)
        
        
        self.backend.run(self)
        
        
        
        qv_qubits = list(qv.reg)
        
        
        type_information = {}
        

        
        if isinstance(qv, QuantumFloat):
            type_information["qtype"] = "qf"
            type_information["msize"] = str(qv.msize)
            type_information["exponent"] = str(qv.exponent)
            type_information["signed"] = str(int(qv.signed))
        elif isinstance(qv, QuantumChar):
            type_information["qtype"] = "qc"
            type_information["nisq_char"] = str(int(qv.nisq_char))
        elif isinstance(qv, QuantumBool):
            type_information["qtype"] = "qb"
            
        
        self.backend.send_qubits(recipient, qv_qubits, str(type_information))
        
        qv.delete()
        
        for i in range(len(qv_qubits)):
            self.qubits.remove(qv_qubits[i])
        
        self.data = []
        
    def update(self):
        """
        This updates the inbox attribute and returns the received QuantumVariables. 

        Returns
        -------
        received_qvs : list[QuantumVariable]
            The QuantumVariables which have been received since the last update.

        Examples
        --------
        
        We assume that the commands from the example of the :meth:`send_qv <qrisp.quantum_network.QuantumNetworkSession.send_qv>` method have been executed.
        
        >>> inbox = bob_session.update()
        >>> print(inbox[0])
        {3.0: 1.0}

        """
        
        from qrisp import QuantumVariable
        
        self.backend.run(self)
        
        updated_inbox = self.backend.inbox()
        
        received_qvs = []
        starting_index = int(len(self.inbox))
        from ast import literal_eval
        
        for i in range(starting_index, len(updated_inbox)):
            
            for qb in updated_inbox[i][0]:
                self.add_qubit(qb)
            
            try:
                type_information = literal_eval(updated_inbox[i][1])
            except:
                type_information = {}
            
                
            
            new_qv = QuantumVariable(0, qs = self)
            
            new_qv.extend(len(updated_inbox[i][0]), proposal = updated_inbox[i][0])
            
            if len(type_information):
                if type_information["qtype"] == "qf":
                    
                    new_qv.__class__ = QuantumFloat
                    new_qv.msize = int(type_information["msize"])
                    new_qv.exponent = int(type_information["exponent"])
                    new_qv.signed = bool(int(type_information["signed"]))
                    
                if type_information["qtype"] == "qc":
                    new_qv.__class__ = QuantumChar
                    new_qv.nisq_char = bool(int(type_information["nisq_char"]))
                    
                elif type_information["qtype"] == "qb":
                    new_qv.__class__ = QuantumBool
                
                    
            
            
            received_qvs.append(new_qv)
            
        self.inbox += received_qvs
        return received_qvs


"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

import numpy as np

#Checks if a gate is permeable on a set of qubits
#Permeable means that the gate has the form
# U = \sum_{x = 0}^{2^n} |x><x| U_x
#where the braket expression represents the qubits
#of the list "qubit_indices"
#{U_x | x <= 2^n} is a set of operations on the remaining qubits

#Applying a gate which is permeable on a register R to a computational basis state in R
#leaves the state of R invariant ie. if R is in state |x>
# U |x>|y> = |x> U_x |y>
#this property will for instance be used to make sure that uncomputations,
#which are depending on the state of R indeed stil recieve the original state of R
#when the uncomputation is performed. An error will be thrown if in the meantime,
#there have been non-permeable operations on R

#But what is the general strategy in determining this property?
#To understand this, note that if we order the computational basis vectors,
#such that all the qubits which are supposed to be checked for permeability are at the end

#In this basis, the unitary is block diagonal, where the blocks on the diagonal are the U_x matrices
#we check the block diagonality by first calculating the block matrices and then create a boolean 2^n x 2^n
#matrix, where the entry is False if the corresponding block matrix is 0

#We then check the diagonality of this matrix and return True if so
def is_permeable(gate, qubit_indices):
    
    for i in qubit_indices:
        if gate.permeability[i] == False:
            return False
        elif gate.permeability[i] is None:
            break
    else:
        return True
    
    from qrisp.circuit import PTControlledOperation
    from qrisp.simulator import calc_embedded_unitary
    
    if gate.definition:
        
        qc = gate.definition
        for instr in qc.data:
            relevant_qubits = set([qc.qubits[k] for k in qubit_indices]).intersection(instr.qubits)
            
            if not relevant_qubits:
                continue
            
            if instr.op.name == "measure":
                return False
            
            local_qubit_indices = [instr.qubits.index(qb) for qb in relevant_qubits]
            
            if not is_permeable(instr.op, local_qubit_indices):
                break
        
        else:
            for i in qubit_indices:
                gate.permeability[i] = True
        
            return True

    #Create qubit order
    qubit_order = []
    for i in range(gate.num_qubits):
        if not i in qubit_indices:
            qubit_order.append(i)
            
    #The qubits which are to be checked come first
    qubit_order = qubit_indices + qubit_order
    
    #Invert permutation
    qubit_order = [qubit_order.index(i) for i in range(len(qubit_order))]
    
    #Calculate unitary
    unitary = calc_embedded_unitary(gate, gate.num_qubits, qubit_order)
    
    if not isinstance(unitary, np.ndarray):
        unitary = unitary.to_array()
    
    #Calculate boolean matrix
    bbm = get_boolean_block_matrix(unitary, 2**(gate.num_qubits - len(qubit_indices)))
    #Check non zero off diagonal elements
    off_diagonal_non_zero_count = np.count_nonzero(bbm - np.diag(np.diagonal(bbm)))
    #Return result
    
    
    from numpy.linalg import norm
    result = not bool(off_diagonal_non_zero_count)
    
    if result:
        for i in qubit_indices:
            gate.permeability[i] = True
    
    
    return result


#Tests if a gate is qfree where qfree is the property that is introduced in SILQ
#meaning that an operation maps any computational basis state to a computational basis state
def is_qfree(gate):
    
    
    if not gate.is_qfree is None:
        return gate.is_qfree
    
    if gate.definition:
        
        for instr in gate.definition.data:
            if not is_qfree(instr.op):
                break
        else:
            gate.is_qfree = True
            return True
    
    unitary = np.around(gate.get_unitary(), 5)
    #Check if the unitary has only one non-zero entry in each line
    for i in range(unitary.shape[0]):
        if np.count_nonzero(unitary[i,:]) != 1:
            gate.is_qfree = False
            return False
        
    gate.is_qfree = True
    return True



    
#Splits the given matrix into blocks of size block_size
#Returns an object array where each entry contains the corresponding block
def get_block_matrix(matrix, block_size):
    
    matrix = np.array(matrix)
    if matrix.shape[0] != matrix.shape[1]:
        raise Exception("Tried to decompose non-square blockmatrix")
        
    if matrix.shape[0]%block_size:
        raise Exception("Matrix shape is no integer multiple of block size")
        
    
    block_amount = int(matrix.shape[0]/block_size)
    result = np.zeros((block_amount, block_amount), dtype = "object")
    
    for i in range(block_amount):
        for j in range(block_amount):
            result[i,j] = matrix[i*block_size:(i+1)*block_size, j*block_size:(j+1)*block_size]
    
    return result


#Recieves a matrix which is turned into a block matrix and then
#returns a matrix of boolean entries where the entry is False, if
#the corresponding block is the 0 matrix and True otherwise
def get_boolean_block_matrix(matrix, block_size):
    
    
    block_matrix = get_block_matrix(matrix, block_size)
        
    block_amount = int(matrix.shape[0]/block_size)
    
    result = np.zeros((block_amount, block_amount))
    
    for i in range(block_amount):
        for j in range(block_amount):
            
            if np.linalg.norm(block_matrix[i,j]) > 1E-4:
                result[i,j] = 1
    
    return result


"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""


import sys

from qrisp.interface.circuit_converter import convert_circuit
from qrisp.interface.thrift_interface.codegen import BackendService


class BackendClient(BackendService.Client):
    """
    This object allows connecting to Qrisp backend servers.
    
    Parameters
    ----------
    socket_ip : string
        The IP address of the socket of the target server.
    port : int
        The port on which the server is listening.

    
    Examples
    --------

    We assume that the example from BackendServer has been executed in the same console.
    
    >>> from qrisp.interface import BackendClient
    >>> example_backend = BackendClient(socket_ip = "127.0.0.1", port = 8080)
    >>> from qrisp import QuantumCircuit
    >>> qc = QuantumCircuit(2)
    >>> qc.h(0)
    >>> qc.cx(0,1)
    >>> qc.measure(qc.qubits)
    >>> example_backend.run(qc, shots = 1000, token = "lorem ipsum")
    lorem ipsum
    {'0 0': 464, '1 1': 536}
        

    
    """

    def __init__(self, socket_ip, port = None):

        from thrift.transport import TSocket, TTransport
        from thrift.protocol import TBinaryProtocol

        if port == None:
            port = 9090
        # Create the transport for the User Interface to the server
        self.transport = TSocket.TSocket(socket_ip, port)
        # Buffering is critical. Raw sockets are very slow
        self.transport = TTransport.TBufferedTransport(self.transport)
        # Wrap in a protocol
        protocol = TBinaryProtocol.TBinaryProtocol(self.transport)
        
        self.socket_ip = socket_ip

        # Create a client to use the protocol encoder
        # client = BackendService.Client(protocol)

        super().__init__(protocol)

        # Connect!
        self.transport.open()

    # Destructor closes the transport
    def __del__(self):
        self.transport.close()

    def run(self, qc, shots, token = ""):
        """
        Executes the ``run_func`` of the server.

        Parameters
        ----------
        qc : QuantumCircuit
            The QuantumCircuit to execute.
        shots : int
            The amount of shots.
        token : str, optional
            A token which can be used for backend execution parameter specification. The default is "".

        Returns
        -------
        dict
            A dictionary representing the counts where the keys are bitstrings and the values are integers.

        """
        

        converted_circuit = convert_circuit(qc, "thrift", transpile = False)
        
        
        return super().run(converted_circuit, shots, token)
    
    def ping(self):
        """
        Executes the ``ping_func`` on the server.
        
        Returns
        -------
        BackendStatus object
            An object which contains general information about the backend.
        """
        
        return super().ping()





"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""


import sympy as sp
from qrisp.logic_synthesis.truth_tables import TruthTable, synth_poly
import math as math



def pprm_synth(input_var, output_var, tt, qb_nr, phase_tolerant = False):
    from qrisp.arithmetic.poly_tools import symbol_to_qbit, monomial_to_list, expr_to_list
    from qrisp import mcx
    qs = input_var.qs
    output_qubit = output_var.reg[qb_nr]
    expr = synth_poly(tt, column = qb_nr)
    #print(expr)
    args = expr_to_list(expr)
    args_len = [len(i) for i in args]
    
    for element in range(len(args_len)):
        
        if args_len[element] > 1:
            from sympy.core.numbers import One
            if isinstance(args[element][0], One):
                args[element].pop(0)
            
            #input_qubits = [(-int(symb.name[1:])-1)%input_var.size for symb in args[element]]
            input_qubits = [int(symb.name[1:]) for symb in args[element]]
            input_qubits = [input_var.reg[i] for i in input_qubits]
            
            product = 1 
            for i in range(len(args[element])):
                product = product * (args[element][i])
            

            mul_tt = TruthTable(product)
            
            if len(product.args) == 0:
                if isinstance(product, sp.Symbol):
                    qs.cx(input_qubits[0], output_qubit)
                else:
                    qs.x(output_qubit)
            else:
                
                if phase_tolerant:
                    qs.append(mul_tt.gate_synth(method = "gray_pt"),input_qubits+[output_qubit])
                else:
                    mcx(input_qubits, output_qubit, method = "gray")
                    # qs.append(mul_tt.gate_synth(method = "gray"),input_qubits+[output_qubit])
        else: 
            if isinstance(args[element][0],sp.core.symbol.Symbol):
                arg_qubit = (-int(args[element][0].name[1:])-1)%input_var.size
                
                arg_qubit = input_var.reg[arg_qubit]
                qs.cx(arg_qubit,output_qubit)
            if isinstance(args[element][0],sp.core.numbers.One):
                qs.x(output_qubit)
   

def pprm(input_var, output_var, tt, phase_tolerant = False):
     for column in range(tt.shape[1]):
        pprm_synth(input_var, output_var, tt, column, phase_tolerant)
        


"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""


def test_quantum_arithmetic():
    
    import numpy as np
    from qrisp import h, multi_measurement, q_div, QuantumFloat
    def test_arithmetic_helper(qf_0, qf_1, operation):
        
        qf_0 = qf_0.duplicate()
        qf_1 = qf_1.duplicate()
        
        h(qf_0)        
        h(qf_1)
        
        if operation == "add":
            qf_res = qf_0 + qf_1
        elif operation == "sub":
            qf_res = qf_0 - qf_1
        elif operation == "mul":
            qf_res = qf_0 * qf_1
        elif operation == "div":
            qf_res = q_div(qf_0, qf_1, prec = 3)

        mes_res = multi_measurement([qf_0, qf_1, qf_res])
        
        for (a, b, c) in mes_res.keys():
            
            if operation == "add":
                assert a + b == c
            elif operation == "sub":
                assert a - b == c
            elif operation == "mul":
                assert a * b == c
            elif operation == "div":
                if b == 0:
                    continue
                assert abs(a / b - c) < 2**(-3)

        statevector = qf_res.qs.statevector("array")
        angles = np.angle(statevector[np.abs(statevector) > 1/2**((qf_0.size + qf_1.size)/2+1)])
        
        #Test correct phase behavior
        assert np.sum(np.abs(angles)) < 0.1
        
    a = QuantumFloat(3, -1, signed = True)
    b = QuantumFloat(5, 1, signed = False)

    for operation in ["add", "sub", "mul", "div"]:
        print(operation)
        test_arithmetic_helper(a, b, operation)

    a = QuantumFloat(3, 1, signed = True)
    b = QuantumFloat(3, 0, signed = True)

    for operation in ["add", "sub", "mul", "div"]:
        test_arithmetic_helper(a, b, operation)

    #Test tensordot
    from qrisp import QuantumFloat, QuantumArray, tensordot, OutcomeArray
    qf = QuantumFloat(3,0, signed = False)
    q_tensor_0 = QuantumArray(qf, shape = (4,4))
    q_tensor_0[:] = np.eye(4)
    q_tensor_0 = q_tensor_0.reshape(2, 2, 2, 2)
    q_tensor_1 = QuantumArray(qf, shape = (2,2))
    q_tensor_1[:] = [[0,1],[1,0]]
    res = tensordot(q_tensor_0, q_tensor_1, (-1, 0))
    
    assert res.get_measurement() == {OutcomeArray([[[[0, 1],
                    [0, 0]],

                   [[1, 0],
                    [0, 0]]],


                  [[[0, 0],
                    [0, 1]],

                   [[0, 0],
                    [1, 0]]]]): 1.0}
    
    #Test inplace multiplication
    b = QuantumFloat(5, signed = True)
    b[:] = 3
    h(b[0])
    h(b[-1])
    b *= -4
    assert b.get_measurement() == {116: 0.25, 120: 0.25, -12: 0.25, -8: 0.25}


"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

# Created by ann81984 at 07.05.2022
import numpy as np
from qrisp import *

from qrisp import QuantumVariable, mcx, cx, QuantumEnvironment, x
def test_uncomputation_example():
    
    verify[0] = 1

    qf1 = QuantumFloat(3, name = "a")

    qf2 = QuantumFloat(3, name = "b")
    qf1[:] = 4

    h(qf1[0])

    qf2[:] = 5

    res = qf1*qf2

    temp = res.duplicate()

    qbool = QuantumBool()
    h(qbool)

    with qbool == True:
        cx(res, temp)

    res.uncompute()
    
    assert multi_measurement([temp, qbool]) == {(0.0, False): 0.5, (20.0, True): 0.25, (25.0, True): 0.25}
    
    #------------------
    print("Test 1 passed")
    

    @auto_uncompute
    def sqrt_oracle(qf):
        z(qf*qf == 0.25)

    qf = QuantumFloat(3, -1, signed = False)
    n = qf.size
    iterations = int((2**n)**0.5) 
    h(qf)

    from qrisp.grover import diffuser
    for i in range(iterations):
        sqrt_oracle(qf)
        diffuser(qf)
        print(qf)
    
    assert qf.get_measurement() == {0.5: 0.9453, 0.0: 0.0078, 1.0: 0.0078, 1.5: 0.0078, 2.0: 0.0078, 2.5: 0.0078, 3.0: 0.0078, 3.5: 0.0078}
    
    #---------
    print("Test 2 passed")
    
    
    city_amount = 4

    distance_matrix = np.array([[0,     0.25,   0.125,  0.5],
                                [0.25,  0,      0.625,  0.375],
                                [0.125, 0.625,  0,      0.75],
                                [0.5,   0.375,  0.75,   0]])/4




    def calc_perm_travel_distance(permutation, precision):
      
        #A QuantumFloat with n qubits and exponent -n 
        #can represent values between 0 and 1
        res = QuantumFloat(precision, -precision)
        
        #Fill QuantumDictionary
        qd = QuantumDictionary(return_type = res)
        for i in range(city_amount):
            for j in range(city_amount):
                qd[(i, j)] = distance_matrix[i, j]
        
        #Evaluate result
        for i in range(4):
            trip_distance = qd[permutation[i], permutation[(i+1)%city_amount]]
            res += trip_distance
            
            trip_distance.uncompute()
        
        
        return res



    

    perm = QuantumArray(QuantumFloat(2))
    perm[:] = [2,1,0,3]

    res = calc_perm_travel_distance(perm, 5)
    
    assert res.get_measurement() == {0.53125: 1.0}
    
    assert len(res.qs.compile().qubits) == 18
    
    #--------
    print("Test 3 passed")
    

    def triple_AND(a, b, c):

      local_quantum_bool = QuantumBool(name = "local")
      result =  QuantumBool(name = "result")
      
      mcx([a, b], local_quantum_bool)
      mcx([local_quantum_bool, c], result)
      
      local_quantum_bool.uncompute()
      
      return result

    a = QuantumBool(name = "a")
    b = QuantumBool(name = "b")
    c = QuantumBool(name = "c")

    h(a)
    b[:] = True
    c[:] = True

    result = triple_AND(a, b, c)

    assert result.get_measurement() == {False: 0.5, True: 0.5}    
    
    #--------

    @auto_uncompute
    def triple_AND(a, b, c):

      local_quantum_bool = QuantumBool(name = "local")
      result =  QuantumBool(name = "result")
      
      mcx([a, b], local_quantum_bool)
      mcx([local_quantum_bool, c], result)
      
      return result

    a = QuantumBool(name = "a")
    b = QuantumBool(name = "b")
    c = QuantumBool(name = "c")

    h(a)
    b[:] = True
    c[:] = True

    result = triple_AND(a, b, c)

    assert result.get_measurement() == {False: 0.5, True: 0.5}

    
    print("Test 4 passed")
    
    #--------
    
    #Test uncomputation/environment interaction
    
    error_thrown = False
    
    d = QuantumVariable(1, name = "d")
    cond_1 = QuantumBool(name = "cond_1")
    cond_2 = QuantumBool(name = "cond_2")
    cond_3 = QuantumBool(name = "cond_3")
    
    
    with cond_1:
        a = QuantumVariable(1, name = "a")    
        b = QuantumVariable(1, name = "b")
        # x(a)
        x(b)
        with cond_2:
            
            x(a)
            x(d)
            with cond_3:
                x(b)
            try:
                a.uncompute()
            except:
                error_thrown = True
        
        
        b.uncompute()
    
    assert error_thrown

    print("Test 5 passed")
    
    #Test recomputation
    
    qv_0 = QuantumVariable(1)
    qv_1 = QuantumVariable(1)
    qv_2 = QuantumVariable(1)

    cx(qv_0, qv_1)
    cx(qv_1, qv_2)

    qv_1.uncompute(recompute = True)

    qv_2.uncompute()
    print(qv_1.qs)

    qv_0 = QuantumVariable(1)
    qv_1 = QuantumVariable(1)
    qv_2 = QuantumVariable(1)
    qv_3 = QuantumVariable(1)

    cx(qv_0, qv_1)
    cx(qv_1, qv_2)
    cx(qv_1, qv_3)
    x(qv_2)

    qv_1.uncompute(recompute = True)
    qv_2.uncompute()
    qv_3.uncompute()
    assert qv_1.qs.cnot_count() == 10
    
    verify[0] = 0


"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""


from qrisp.interface.circuit_converter import convert_circuit
from qrisp.quantum_network.interface.codegen import QuantumNetworkService
from qrisp.quantum_network.interface.codegen.ttypes import Message
from qrisp.interface import PortableQubit
from qrisp import Qubit

class QuantumNetworkClient(QuantumNetworkService.Client):
    """
    This class allows to connect to Qrisp quantum networks.
    
    Parameters
    ----------
    name : str
        A nickname such as "Alice" or "Bob", which is used to direct the communication of other users to you.
    socket_ip : str
        The IP adress of the QuantumNetworkServer.
    port : int, optional
        The port of the QuantumNetworkServer. The default is 7070.
        
    Examples
    --------
    
    We create a QuantumNetworkServer listening on the localhost IP-address an connect the client.
    
    >>> from qrisp.quantum_network import QuantumNetworkServer, QuantumNetworkClient
    >>> local_server = QuantumNetworkServer("127.0.0.1", background = True)
    >>> local_server.start()
    >>> client = QuantumNetworkClient(name = "alice", socket_ip = "127.0.0.1")
    
    
    """
    
    
    def __init__(self, name, socket_ip, port = None):
        
        from thrift.transport import TSocket, TTransport
        from thrift.protocol import TBinaryProtocol
        
        if port == None:
            port = 7070
        #Create the transport for the User Interface to the server
        self.transport = TSocket.TSocket(socket_ip, port)
        # Buffering is critical. Raw sockets are very slow
        self.transport = TTransport.TBufferedTransport(self.transport)
        # Wrap in a protocol
        protocol = TBinaryProtocol.TBinaryProtocol(self.transport)
        self.socket_ip = socket_ip        
        # Create a client to use the protocol encoder
                        
        super().__init__(protocol)
        
        # Connect!
        self.transport.open()
        
        
        self.name = name
        self.register_client(name)
    
    #Destructor closes the transport
    def __del__(self):
    
        self.transport.close()
        
    def run(self, qc, shots = None):
        r"""
        Runs a QuantumCircuit on the client's backend. Note that QuantumNetwork simulations do not support multiple shots, as the quantum state of the network is stored and updated everytime a client sends a query. Multiple shots could yield differing measurement outcomes, which implies an ambiguous quantum state of the network. Nevertheless the results are returned in the form of a dictionary in order to comply with the quantum circuit execution backend interface.
        
        Note that it is possible to submit QuantumCircuits which contain qubits, that have not been requested previously. In this case, the qubits names are internally extended by the string "@client_name"  (if they aren't extended in this way already). This is to allow multiple clients to submit circuits with matching qubit names.

        Parameters
        ----------
        qc : QuantumCircuit
            The QuantumCircuit to run.

        Returns
        -------
        res : dict
            A dictionary containing a single key/value pair where the key represents the measurement outcome.
            
        
        Examples
        --------
        
        We create a local QuantumNetworkServer, connect a client and run a QuantumCircuit
        
        >>> from qrisp.quantum_network import QuantumNetworkServer, QuantumNetworkClient
        >>> local_server = QuantumNetworkServer("127.0.0.1", background = True)
        >>> local_server.start()
        >>> client = QuantumNetworkClient(name = "alice", socket_ip = "127.0.0.1")
        >>> from qrisp import QuantumCircuit
        >>> qc = QuantumCircuit(2)
        >>> qc.h(0)
        >>> qc.cx(0,1)
        >>> qc.measure([0,1])
        >>> client.run(qc)
        {'11': 1}
        
        We perform another shot
        
        >>> client.run(qc)
        {'01': 1}

        After applying the first run command, the quantum state is

        .. math::        
            \ket{\psi} = \ket{1}\ket{1}
        
        Applying the Hadamard yields
        
        .. math::        
            \text{H}_0 \ket{\psi} = \frac{1}{\sqrt{2}}(\ket{0} - \ket{1})\ket{1}
        
        The CX yields
        
        .. math::        
            \text{CX}_{01} \text{H}_0 \ket{\psi} = \frac{1}{\sqrt{2}}(\ket{0}\ket{1} - \ket{1}\ket{0})
        
        Finally, the measurement collapsed the state into the first summand.
        
        """
        converted_circuit = convert_circuit(qc, "thrift")
        
        res = super().run(converted_circuit, self.name)
        
        return res
    

        
        qc = super().get_clear_qc(self.name)
        
        return convert_circuit(qc, "qrisp")

    def send_qubits(self, recipient, qubits, annotation):
        """
        Sends the specified qubits to another participant of the network.

        Parameters
        ----------
        recipient : str
            The recipients name.
        qubits : list[Qubit]
            The list of qubits to send.
        annotation : str
            A string containing an arbitrary message that is available for the reciepient.

        
        Examples
        --------
        
        We create a QuantumNetworkServer, connect two clients and distribute a bell pair another client.
        
        >>> from qrisp.quantum_network import QuantumNetworkServer, QuantumNetworkClient
        >>> local_server = QuantumNetworkServer("127.0.0.1", background = True)
        >>> local_server.start()
        >>> alice_client = QuantumNetworkClient(name = "alice", socket_ip = "127.0.0.1")
        >>> bob_client = QuantumNetworkClient(name = "bob", socket_ip = "127.0.0.1")
        
        Now we create the bell pair:
        
        >>> from qrisp import QuantumCircuit
        >>> qc = QuantumCircuit(2)
        >>> qc.h(0)
        >>> qc.cx(0,1)
        >>> alice_client.run(qc)
        {'': 1}
        
        Send one of the qubits to Bob:
            
        >>> alice_client.send_qubits("bob", [qc.qubits[0]], annotation = "Happy birthday!")
        
        Now Bob can measure
        
        >>> messages = bob_client.inbox()
        >>> received_qubits, annotation = messages[0]
        >>> qc = bob_client.get_clear_qc()
        >>> qc.measure(received_qubits)
        >>> bob_client.run(qc)
        {'1': 1}
        
        After Bob's measurement, we expect Alice's measurement to yield the same due to `spukhafter Fernwirkung <https://en.wikipedia.org/wiki/Quantum_entanglement#History>`_:
            
        >>> qc = alice_client.get_clear_qc()
        >>> qc.measure(qc.qubits)
        >>> alice_client.run(qc)
        {'1': 1}
        


        """
        
        msg = Message([PortableQubit(qb.identifier) for qb in qubits], annotation)
        
        super().send_qubits(self.name, recipient, msg)

    
    def request_qubits(self, amount):
        """
        Creates the specified amount of qubits for the client's backend.

        Parameters
        ----------
        amount : int
            The amount of qubits to create.

        Returns
        -------
        list[Qubit]
            The qubits created.
            
        Examples
        --------

        >>> from qrisp.quantum_network import QuantumNetworkServer, QuantumNetworkClient
        >>> local_server = QuantumNetworkServer("127.0.0.1", background = True)
        >>> local_server.start()
        >>> client = QuantumNetworkClient(name = "alice", socket_ip = "127.0.0.1")            
        >>> qb_list = client.request_qubits(4)
        >>> print(qb_list)        
        [Qubit(qb_0@alice), Qubit(qb_1@alice), Qubit(qb_2@alice), Qubit(qb_3@alice)]
        
        """
        
        qb_list = super().request_qubits(amount, self.name)
        
        return [Qubit(qb.identifier) for qb in qb_list]

    def get_clear_qc(self):
        """
        Returns a QuantumCircuit containing all the qubits that belong to the client at the moment.
        
        Note that the qubit names of the run QuantumCircuit are internally extended by the string "@client_name"  (if they aren't extended in this way already). This is to allow multiple clients to submit circuits with matching qubit names.
        
        Note that 

        Returns
        -------
        QuantumCircuit
            An empty QuantumCircuit containing the qubits, this client may operate on.
            
        Examples
        --------
        
        We set up a quantum network, run some QuantumCircuits and retrieve the clear QuantumCircuit:

        >>> from qrisp.quantum_network import QuantumNetworkServer, QuantumNetworkClient
        >>> local_server = QuantumNetworkServer("127.0.0.1", background = True)
        >>> local_server.start()
        >>> client = QuantumNetworkClient(name = "alice", socket_ip = "127.0.0.1")            
        >>> from qrisp import QuantumCircuit
        >>> qc_0 = QuantumCircuit(1)
        >>> qc_0.x(0)
        >>> print(qc_0)
              ┌───┐
        qb_8: ┤ X ├
              └───┘
        >>> client.run(qc_0)
        >>> qc_1 = QuantumCircuit(1)
        >>> qc_1.h(0)
        >>> print(qc_1)
                ┌───┐
         qb_26: ┤ H ├
                └───┘
        >>> client.run(qc_1)
        >>> print(client.get_clear_qc().qubits)
        [Qubit(qb_58@alice), Qubit(qb_77@alice)]
        
        """
        qc = super().get_clear_qc(self.name)
        
        return convert_circuit(qc, "qrisp")


    def get_overall_qc(self):
        """
        Retrieves the QuantumCircuit of all operations that happened in the network so far.

        Returns
        -------
        QuantumCircuit
            The QuantumCircuit containing all operations that happened in this QuantumNetwork.

        Examples
        --------
        
        We assume that the commands of the example of the :meth:`send_qubits <qrisp.quantum_network.QuantumNetworkClient.send_qubits>` method have been executed.
        
        >>> print(alice_client.get_overall_qc())
                      ░ ┌───┐      ░ ┌─┐ ░    
        qb_34@alice: ─░─┤ H ├──■───░─┤M├─░────
                      ░ └───┘┌─┴─┐ ░ └╥┘ ░ ┌─┐
        qb_35@alice: ─░──────┤ X ├─░──╫──░─┤M├
                      ░      └───┘ ░  ║  ░ └╥┘
           cb_bob_0: ═════════════════╩═════╬═
                                            ║ 
         cb_alice_0: ═══════════════════════╩═
                                              
                                    
        """
        
        
        qc = super().get_overall_qc()
        
        return convert_circuit(qc, "qrisp")
    
    def inbox(self):
        """
        Returns a list of tuples containing the received qubits and their annotations.

        Returns
        -------
        res_list : list[tuple[list[Qubit], str]]
            A list of tuples containing the send qubits and the annotations.
        
        
        Examples
        --------
        
        We send a qubit from one client to another and inspect the inbox
        
        >>> from qrisp.quantum_network import QuantumNetworkServer, QuantumNetworkClient
        >>> local_server = QuantumNetworkServer("127.0.0.1", background = True)
        >>> local_server.start()
        >>> alice_client = QuantumNetworkClient(name = "alice", socket_ip = "127.0.0.1")
        >>> bob_client = QuantumNetworkClient(name = "bob", socket_ip = "127.0.0.1")
        
        Prepare the qubit which will be sent to bob:
        
        >>> from qrisp import QuantumCircuit
        >>> qc = QuantumCircuit(1)
        >>> qc.h(0)
        >>> alice_client.run(qc)
        {'': 1}
        
        Send the qubit to Bob:
            
        >>> alice_client.send_qubits("bob", [qc.qubits[0]], annotation = "Merry christmas!")
        
        Now Bob can check his inbox
        
        >>> messages = bob_client.inbox()
        >>> received_qubits, annotation = messages[0]
        >>> print(received_qubits)
        [Qubit(qb_20@alice)]
        >>> print(annotation)
        Merry christmas!

        """
        
        inbox_list = super().inbox(self.name)
        
        res_list = []
        
        for i in range(len(inbox_list)):
            
            res_list.append(([Qubit(qb.identifier) for qb in inbox_list[i].qubits], inbox_list[i].annotation))
        return res_list           



"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

# -*- coding: utf-8 -*-

import numpy as np
from qrisp import QuantumCircuit, RYGate, QuantumBool, cx, x, h, s, s_dg, mcx, invert, bin_rep, QuantumVariable, TruthTable, XGate, PGate, QuantumSession
from qrisp.misc.GMS_tools import GXX_wrapper

#Interface function to quickly change between different implementations of multi controlled not gates
def multi_cx(n, method = None):
    
    #from qrisp.circuit import transpile
    
    if method == "gms":
        return gms_multi_cx(n)    
        

    elif method == "gray_pt":
        
        return pt_multi_cx(n)
    
    elif method == "gray_pt_inv":
        
        return pt_multi_cx(n).inverse()
        
    elif method in ["gray", "auto", None]:
        
        return gray_multi_cx(n)
    
    else:
        raise Exception("method \"" + method + "\" not implemented")


#Function to synthesize a multi controlled CX gate from GMS gates
def gms_multi_cx(n):
    
    qc = QuantumCircuit(n+1)
    
    if n == 2:
        for i in range(3):
            qc.ry(np.pi/2, i)
        
        qc.rz(np.pi/4, 2)
        
        qc.append(GXX_wrapper(3, 3*[3*[np.pi/2]]), qc.qubits)
        
        for i in range(3):
            qc.rx(-np.pi/2, i)
            
        qc.rz(-np.pi/2, 2)
        
        for i in range(3):
            qc.rx(-np.pi/4, i)
        
        qc.append(GXX_wrapper(3, 3*[3*[np.pi/4]]), qc.qubits)
        
        qc.rz(np.pi/2, 2)

        qc.append(GXX_wrapper(3, 3*[3*[np.pi/2]]), qc.qubits)
        
        for i in range(3):
            qc.rx(np.pi/2,i)
            qc.ry(-np.pi/2,i)
        
        result = qc.to_gate()
        
        result.name = "GMS toffoli"
        
        return result
    else:
        raise Exception(str(n) + "-controlled x gate not implemented for gms method")
        

def gray_pt_mcx(n, ctrl_state):
    
    input_qv = QuantumVariable(n)
    output_qv = QuantumVariable(1, qs = input_qv.qs)
    
    
    tt_str = 2**n*["0"]
    tt_str[int(ctrl_state[::-1], 2)] = "1"
    
    tt = TruthTable([tt_str])
    
    tt.q_synth(input_qv, output_qv, method = "gray_pt")

    res = input_qv.qs.copy()
    return res.to_gate(f"pt{n}cx")

#Function to synthesize a phase tolerant multi controlled X gate
def pt_multi_cx(n, reduced = False):
    
    res = QuantumCircuit(n+1)

    if n == 1:
        res.cx(0,1)
        
    #The special cases n = 2 and n = 3 are handled as described in https://arxiv.org/pdf/1508.03273.pdf
    elif n == 2:
        
        if reduced:
            res.append(reduced_margolus_qc.to_gate("reduced margolus"), res.qubits)
        else:
            res.append(margolus_qc.to_gate("margolus"), res.qubits)
    
    
    elif n == 3:
        
        if reduced:
            res.append(reduced_maslov_qc.to_gate("reduced maslov"), res.qubits)
        else:
            res.append(maslov_qc.to_gate("maslov"), res.qubits)
    else:
        
        input_qv = QuantumVariable(n)
        output_qv = QuantumVariable(1, qs = input_qv.qs)
        
        tt = TruthTable([(2**(n)-1)*"0"+"1"])
        
        tt.q_synth(input_qv, output_qv, method = "gray_pt")
    
        res = input_qv.qs.copy()
        
    return res.to_gate(f"pt{n}cx")


#Toffoli Implementation according to https://arxiv.org/pdf/1206.0758.pdf
toffoli_qc = QuantumCircuit(3)

toffoli_qc.h(2)
toffoli_qc.p(-np.pi/4, [0,1])
toffoli_qc.cx(2,0)
toffoli_qc.cx(1,2)
toffoli_qc.p(np.pi/4, 0)
toffoli_qc.cx(1,0)
toffoli_qc.p(np.pi/4, 2)
toffoli_qc.cx(1,2)
toffoli_qc.p(-np.pi/4, 0)
toffoli_qc.cx(2,0)
toffoli_qc.p(np.pi/4, 0)
toffoli_qc.p(-np.pi/4, 2)
toffoli_qc.cx(1,0)
toffoli_qc.h(2)


margolus_qc = QuantumCircuit(3)
G = RYGate(np.pi/4)

margolus_qc.append(G, 2)
margolus_qc.cx(1,2)
margolus_qc.append(G, 2)
margolus_qc.cx(0,2)

reduced_margolus_qc = margolus_qc.copy()

margolus_qc.append(G.inverse(), 2)
margolus_qc.cx(1,2)
margolus_qc.append(G.inverse(), 2)



maslov_qc = QuantumCircuit(4)

maslov_qc.h(3)
maslov_qc.t(3)        
maslov_qc.cx(2,3)
maslov_qc.t_dg(3)        
maslov_qc.h(3)
maslov_qc.cx(0,3)
maslov_qc.t(3)
maslov_qc.cx(1,3)
maslov_qc.t_dg(3)
maslov_qc.cx(0,3)

reduced_maslov_qc = maslov_qc.copy()

maslov_qc.t(3)
maslov_qc.cx(1,3)
maslov_qc.t_dg(3)
maslov_qc.h(3)
maslov_qc.t(3)
maslov_qc.cx(2,3)
maslov_qc.t_dg(3)
maslov_qc.h(3)


#Ancilla supported multi controlled X gates from https://arxiv.org/pdf/1508.03273.pdf
def maslov_mcx(n, ctrl_state = -1):
    if not isinstance(ctrl_state, str):
        if ctrl_state == -1:
            ctrl_state += 2**n
        ctrl_state = bin_rep(ctrl_state, n)
    
    res = QuantumCircuit(n+1)
    for i in range(len(ctrl_state)):
        if ctrl_state[i] == "0":
            res.x(i)
    
    if n == 1:
        res.cx(0,1)
    elif n == 2:
        res.append(toffoli_qc.to_gate("toffoli"), res.qubits)
    elif n == 3:
        res.add_qubit()
        res.append(margolus_qc.to_gate(), [0,1,3])
        res.append(toffoli_qc.to_gate(), [2,3,4])
        res.append(margolus_qc.inverse().to_gate(), [0,1,3])
    elif n == 4:
        res.add_qubit()
        res.append(maslov_qc.to_gate(), [0,1,2,4])
        res.append(toffoli_qc.to_gate(), [3,4,5])
        res.append(maslov_qc.inverse().to_gate(), [0,1,2,4])
    else:
        raise Exception("Multi CX for method \"Maslov\" only defined for n <= 4")
    
    for i in range(len(ctrl_state)):
        if ctrl_state[i] == "0":
            res.x(i)
            
    return res.to_gate(f"maslov {n}cx")



#Function to synthesize a multi controlled X gate using gray synthesis
def gray_multi_cx(n):
    qs = QuantumSession()
    input_qv = QuantumVariable(n, qs)
    output_qv = QuantumVariable(1, qs)

    if n == 1:
        qs.cx(0,1)
    elif n == 2:
        qs = toffoli_qc.copy()
        
    else:
        tt_array = np.zeros((2**n,1))
        tt_array[-1,0] = 1
        
        tt = TruthTable(tt_array)
        
        tt.q_synth(input_qv, output_qv, method = "gray")
    
    
    result = qs.to_gate()
    result.name = "gray multi cx"
    
    return result


#Ancilla supported multi controlled X with logarithmic depth based on https://www.iccs-meeting.org/archive/iccs2022/papers/133530169.pdf
def balauca_mcx(input_qubits, target, ctrl_state = None, phase = None):
    hybrid_mcx(input_qubits, target, ctrl_state = ctrl_state, phase = phase, num_ancilla = np.inf)


#Algorithm based on https://link.springer.com/article/10.1007/s10773-017-3389-4
def yong_mcx(input_qubits, target, ancilla = None, ctrl_state = None):
    
    if isinstance(target, list):
        if len(target) != 1:
            raise Exception("Tried to execute yong algorithm with multiple targets")
        target = target[0]
    
    if not ctrl_state is None:
        for i in range(len(input_qubits)):
            input_qubits[i].ctrl_state = ctrl_state[i]
    
    if len(input_qubits) < 3:
        ctrl_state = ""
        for i in range(len(input_qubits)):
            if hasattr(input_qubits[i], "ctrl_state"):
                ctrl_state += input_qubits[i].ctrl_state
            else:
                ctrl_state += "1"
    
    
    if len(input_qubits) == 2:
        if ancilla is None:
            mcx(input_qubits, target, method = "gray", ctrl_state = ctrl_state)
        else:
            target.qs().append(gray_pt_mcx(2, ctrl_state = ctrl_state), input_qubits + [target])
            # mcx(input_qubits, target, method = "gray_pt", ctrl_state = ctrl_state)
        return
    
    ancilla_allocated = False
    
    if ancilla is None:
        ancilla_allocated = True
        ancilla_bl = QuantumBool(name = "yong_anc*")
        ancilla = ancilla_bl[0]
    
    n = len(input_qubits)
        
    partition_k_1 = input_qubits[n//2:]
    partition_k_2 = input_qubits[:n//2]
    
    h(target)
    
    yong_mcx(input_qubits = partition_k_1, 
             target = ancilla, 
             ancilla = partition_k_2[-1])
    
    s(ancilla)
    
    yong_mcx(input_qubits = partition_k_2 + [target], 
             target = ancilla, 
             ancilla = partition_k_1[-1])
    
    s_dg(ancilla)
    
    with invert():
        
        s(ancilla)
        
        yong_mcx(input_qubits = partition_k_2 + [target], 
                 target = ancilla, 
                 ancilla = partition_k_1[0])
        
        s_dg(ancilla)
        
        yong_mcx(input_qubits = partition_k_1, 
                 target = ancilla, 
                 ancilla = partition_k_2[0])
    
    h(target)
    
    if ancilla_allocated:
        ancilla_bl.delete()
    
    if not ctrl_state is None:
        for i in range(len(input_qubits)):
            del input_qubits[i].ctrl_state



#Hybrid algorithm of yong and balauca with customizable ancilla qubit count.
#Performs several balauca layers and cancels the recursion with yong.
def hybrid_mcx(input_qubits, target, ctrl_state = None, phase = None, num_ancilla = np.inf, num_dirty_ancilla = 0):
    """
    Function to dynamically generate mcx gates for a given amount of ancilla qubits.

    Parameters
    ----------
    input_qubits : list[Qubit]
        The Qubits to control on.
    target : Qubit
        The Qubit to target.
    ctrl_state : str, optional
        The control state to activate the X Gate on. The default is "11..".
    phase : float or sympy.Symbol, optional
        If given, this function performs a mcp instead of an mcx. The default is None.
    num_ancilla : int, optional
        The amount of ancillae this function is allowed to use. The default is np.inf.
    count_yong : bool, optional
        If set to False, the ancilla that is used by the yong recursion termintation is not counted (because it can be dirty). The default is True.

    Returns
    -------
    None.

    """
    
    input_qubits = list(input_qubits)
    for i in range(len(input_qubits)):
        if isinstance(input_qubits[i], QuantumBool):
            input_qubits[i] = input_qubits[i][0]
    
    if not ctrl_state is None:
        for i in range(len(input_qubits)):
            input_qubits[i].ctrl_state = ctrl_state[i]
            
    target = list(target)
    
    qs = target[0].qs()
    
    if len(input_qubits) <= 3 or num_ancilla == 0:
        ctrl_state = ""
        for qb in input_qubits:
            if hasattr(qb, "ctrl_state"):
                ctrl_state += qb.ctrl_state
            else:
                ctrl_state += "1"
        
        if len(input_qubits) == 2:
            if phase is None:
                qs.append(XGate().control(2, ctrl_state = ctrl_state, method = "gray"), input_qubits + [target])
            else:
                qs.append(PGate(phase).control(2, ctrl_state = ctrl_state), input_qubits + [target])

        elif num_dirty_ancilla and phase is None:
            balauca_dirty(input_qubits, target, k = num_dirty_ancilla, ctrl_state = ctrl_state)
            
        else:
            if phase is None:
                qs.append(XGate().control(len(input_qubits), ctrl_state = ctrl_state, method = "gray"), input_qubits + [target])
            else:
                qs.append(PGate(phase).control(len(input_qubits), ctrl_state = ctrl_state), input_qubits + [target])
        return
    
    structure = structure_decider(len(input_qubits), num_ancilla)#[::-1]
    layer_input = []
    layer_structure = []
    layer_output = []
    remainder = list(input_qubits)
    
    
    for i in range(len(structure)):
        if not structure[0] <= len(remainder):
            break
            
        layer_output.append(QuantumBool(name = "balauca_anc*"))
        
        for j in range(structure[0]):
            layer_input.append(remainder.pop(0))

        layer_structure.append(structure.pop(0))
    
    
    balauca_layer(layer_input, layer_output, structure = layer_structure, invert = False)

    hybrid_mcx(layer_output + remainder, target, num_ancilla = num_ancilla - len(layer_output), num_dirty_ancilla = num_dirty_ancilla, phase = phase)

    balauca_layer(layer_input, layer_output, structure = layer_structure, invert = True)
    
    
    [qbl.delete() for qbl in layer_output]
    
    if not ctrl_state is None:
        for i in range(len(input_qubits)):
            del input_qubits[i].ctrl_state
    
    return
    

def balauca_layer(input_qubits, output_qubits, structure, invert = False):
    
    if not output_qubits:
        return
    
    qs = output_qubits[0].qs()
    input_qubits = list(input_qubits)
    
    counter = 0
    for i in range(len(output_qubits)):
        
        if structure[i] == 3:
            
            ctrl_qubits = [input_qubits[counter], input_qubits[counter+1], input_qubits[counter+2]]
            counter += 3
            
            ctrl_state = ""
            for qb in ctrl_qubits:
                
                if hasattr(qb, "ctrl_state"):
                    ctrl_state += qb.ctrl_state
                else:
                    ctrl_state += "1"
            
            
            gate = XGate().control(3, method = "gray_pt", ctrl_state = ctrl_state)
            
            if invert:
                gate = gate.inverse()
                
            qs.append(gate, ctrl_qubits + [output_qubits[i]])
        else:
            
            ctrl_qubits = [input_qubits[counter], input_qubits[counter+1]]
            counter += 2
            
            ctrl_state = ""
            for qb in ctrl_qubits:
                
                if hasattr(qb, "ctrl_state"):
                    ctrl_state += qb.ctrl_state
                else:
                    ctrl_state += "1"
            
            
            
            gate = XGate().control(2, method = "gray_pt", ctrl_state = ctrl_state)
            
            if invert:
                gate = gate.inverse()
            
            qs.append(gate, ctrl_qubits + [output_qubits[i]])
            
    return

def structure_decider(n, k):
    #Each element of a Balauca layer reduces the amount of
    #Control qubits for the next layer either by 1 (margolous gate)
    #or by 2 (phase toleratn maslov gate). Ideally, we reduce the amount
    #of control qubits to 3 so we can cancel the recursion
    #and deploy a Toffoli.
    
    #If we have n controls, p = n - 2 is the amount of reductions,
    #that need to be performed.
    
    #If we have k ancillae at our disposal, we need to satisfy:
        
    #p = 2*triple_mcx + (k-triple_mcx_count) = k + triple_mcx_count
    
    # <=> triple_mcx_count = n - 2 - k
    
    triple_mcx_count = n - 2 - k
    
    if triple_mcx_count <= 0:
        if n%2:
            return (n-3)//2*[2] + [3]
        else:
            return n//2*[2]
    elif triple_mcx_count > k:
        return k*[3]
    else:
        return triple_mcx_count*[3] + (k-triple_mcx_count)*[2]
    
    
    
from qrisp import convert_to_qb_list
def vchain_2_dirty(control, target, dirty_ancillae = None):
    
    
    control = convert_to_qb_list(control)
    target = convert_to_qb_list(target)
    
    
    if len(control) == 1:
        cx(control, target)
        return
    elif len(control) == 2:
        mcx(control, target, method = "gray")
    
    
    n = len(control)
    k = n - 2
    
    dirty_ancilla_qbls = []
    if dirty_ancillae is None:
        dirty_ancilla_qbls = [QuantumBool(name = "vchain_2_dirty*") for i in range(k)]
        dirty_ancillae = [qbl[0] for qbl in dirty_ancilla_qbls]
    
    def reduced_margolus(control, target):
        qs = target.qs()
        
        control = list(control)
        qs.append(reduced_margolus_qc.to_gate("reduced_margolus"), control + [target])
        # qs.append(XGate().control(2), control + [target])
    
    def margolus(control, target):
        qs = target.qs()
        
        control = list(control)
        qs.append(margolus_qc.to_gate("margolus"), control + [target])
        # qs.append(XGate().control(2), control + [target])
    
    
    control_temp_list = list(control)
    dirty_ancillae_temp_list = list(dirty_ancillae)
    qubit_list = []
    
    qubit_list.append(control_temp_list.pop(0))
    qubit_list.append(control_temp_list.pop(0))
    
    while control_temp_list:
        qubit_list.append(dirty_ancillae_temp_list.pop(0))
        qubit_list.append(control_temp_list.pop(0))
    
    
    qubit_list.append(target[0])
    
    m = len(qubit_list)
    
    for i in range(k):
        if i == k - 1:
            margolus([qubit_list[m-5-2*i], qubit_list[m-4-2*i]], qubit_list[m-3-2*i])
        else:
            reduced_margolus([qubit_list[m-5-2*i], qubit_list[m-4-2*i]], qubit_list[m-3-2*i])
    
    
    for i in range(k):
        if i == k - 1:
            reduced_margolus([qubit_list[2*i+2], qubit_list[2*i+3]], qubit_list[2*i+4])
        else:
            with invert():
                reduced_margolus([qubit_list[2*i+2], qubit_list[2*i+3]], qubit_list[2*i+4])
    
    
    for i in range(k):
        if i == k - 1:
            with invert():
                margolus([qubit_list[m-5-2*i], qubit_list[m-4-2*i]], qubit_list[m-3-2*i])
        else:
            reduced_margolus([qubit_list[m-5-2*i], qubit_list[m-4-2*i]], qubit_list[m-3-2*i])
    
    
    for i in range(k):
        if i == k - 1:
            with invert():
                reduced_margolus([qubit_list[2*i+2], qubit_list[2*i+3]], qubit_list[2*i+4])
        else:
            with invert():
                reduced_margolus([qubit_list[2*i+2], qubit_list[2*i+3]], qubit_list[2*i+4])        


    [qbl.delete() for qbl in dirty_ancilla_qbls]
    

def balauca_dirty(control, target, k, dirty_ancillae = None, ctrl_state = None):
    
    control = convert_to_qb_list(control)
    target = convert_to_qb_list(target)
    qs = target[0].qs()
    
    n = len(control)
    
    k = min((n-2)//2+1, k)
    
    if k == 0:
        qs.append(XGate().control(n, ctrl_state = ctrl_state, method = "gray"), list(control) + [target])
        return
    
    
    if not ctrl_state is None:
        for i in range(len(ctrl_state)):
            if ctrl_state[i] == "0":
                x(control[i])
    
    
    def reduced_maslov(control, target):
        qs = target.qs()
        
        qs.append(reduced_maslov_qc.to_gate("reduced_maslov"), control + [target])
        # qs.append(XGate().control(2), control + [target])
    
    
    m_1 = int(np.ceil((n - 2*(k-1))/2))
    m_2 = int(np.floor((n - 2*(k-1))/2))
    
    dirty_ancilla_qbls = []
    if dirty_ancillae is None:
        dirty_ancilla_qbls = [QuantumBool(name = "balauca_dirty*", qs = qs) for i in range(k)]
        dirty_ancillae = [qbl[0] for qbl in dirty_ancilla_qbls]
    
    
    upper_block_qubits = control[:m_1] + [dirty_ancillae[0]]
    lower_block_qubits = control[-m_2:] + [dirty_ancillae[-1]]
    
    def balauca_dirty_helper(control, target, ancillae):
        
        if len(ancillae) == 0:
            vchain_2_dirty(control, target, dirty_ancillae = lower_block_qubits)
            # mcx(control, target)
            return
        
        reduced_maslov([ancillae[-1]] + control[-2:], target)
        
        balauca_dirty_helper(control[:-2], ancillae[-1], ancillae[:-1])
        
        with invert():
            reduced_maslov([ancillae[-1]] + control[-2:], target)
            
    
    vchain_2_dirty(lower_block_qubits, target, dirty_ancillae = upper_block_qubits)
    
    balauca_dirty_helper(control[:-m_2], dirty_ancillae[-1], dirty_ancillae[:-1])
    
    vchain_2_dirty(lower_block_qubits, target, dirty_ancillae = upper_block_qubits)
    
    balauca_dirty_helper(control[:-m_2], dirty_ancillae[-1], dirty_ancillae[:-1])

    [qbl.delete() for qbl in dirty_ancilla_qbls]
    
    if not ctrl_state is None:
        for i in range(len(ctrl_state)):
            if ctrl_state[i] == "0":
                x(control[i])



"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""


from qrisp.interface import BackendServer, BackendClient
from qrisp import QuantumVariable, QuantumCircuit, h, x, cx

def test_interface():
    def run_func(qc, shots, token):
        
        print("Recieved query")
        
        #Convert to qiskit
        from qrisp.interface.circuit_converter import convert_circuit
        
        qiskit_qc = convert_circuit(qc, "qiskit")
        
        print(qiskit_qc)
        
        from qiskit import Aer
        qiskit_backend = Aer.get_backend('qasm_simulator')
        
        #Run Circuit on the Qiskit backend
        return qiskit_backend.run(qiskit_qc, shots = shots).result().get_counts()
    
    
    socket_ip = "::1"
    port = 8081
    
    test_server = BackendServer(run_func, socket_ip_address = socket_ip, port = port)
    
    test_server.start()
    
    
    qv1 = QuantumVariable(2)
    qv2 = QuantumVariable(2)
    
    h(qv1[0])
    x(qv1[1])
    cx(qv1[0], qv2[1])
    
    
    test_client = BackendClient(socket_ip, port)
    
    
    res = qv1.get_measurement(backend = test_client)
    assert "01" in res.keys() and "11" in res.keys()


"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""




from qrisp.interface.openapi_interface.codegen.client.openapi_client.api.default_api import DefaultApi
from qrisp.interface.openapi_interface.codegen.client.openapi_client.models import InlineObject
from qrisp.interface.openapi_interface.codegen.client.openapi_client import Configuration, ApiClient

from qrisp.interface.circuit_converter import convert_circuit


class BackendClient(DefaultApi):
    
    def __init__(self, socket_ip, port):
        
        if socket_ip.find(":") != -1:
            socket_ip = "[" + socket_ip + "]"
        
        if port == None:
            port = 8080
        
        #TO-DO Allow API token/Secure connections ... check Configuration class definition
        config = Configuration(host = "http://" + socket_ip + ":" + str(port))
        client = ApiClient(configuration = config)
        
        super().__init__(client)
    
    
    def run(self, qc, shots, token = ""):
        request_object = InlineObject(qc  = convert_circuit(qc, "open_api"), shots = shots, token = token)
        return super().run(inline_object = request_object)



"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

def test_QuantumArray_indexing():
    
    from qrisp import QuantumBool, QuantumArray, QuantumFloat, h, multi_measurement
    
    q_array = QuantumArray(QuantumBool(), shape = (4,4))
    index_0 = QuantumFloat(2)
    index_1 = QuantumFloat(2)
    
    index_0[:] = 2
    index_1[:] = 1
    
    h(index_0[0])

    with q_array[index_0, index_1] as entry:
        entry.flip()

    assert q_array.most_likely()[index_0.most_likely(), index_1.most_likely()] == True

def test_QuantumDictionary_load():
    
    from qrisp import QuantumDictionary, QuantumFloat, h
    qtype = QuantumFloat(4, -2)
    float_qd = QuantumDictionary(return_type = qtype)
    float_qd[0] = 1
    float_qd[1] = 2
    
    key_qv = QuantumFloat(1, signed = False)
    h(key_qv)
    value_qv = qtype.duplicate()
    
    float_qd.load(key_qv, value_qv, synth_method = "pprm")
    assert value_qv.get_measurement() == {1.0: 0.5, 2.0: 0.5}


def test_QuantumFloat():
    
    from qrisp import QuantumFloat, invert

    qf_6 = QuantumFloat(3)
    qf_7 = QuantumFloat(3)
    qf_6[:] = 7
    qf_7[:] = 2
    qf_8 = qf_6/qf_7
    assert qf_8.get_measurement() == {3.5: 1.0}
    
    qf_9 = qf_6//qf_7
    assert qf_9.get_measurement() == {3.0: 1.0}
    
    qf_10 = QuantumFloat(3, -1)
    qf_10[:] = 3.5
    qf_11 = qf_10**-1
    assert qf_11.get_measurement() == {0.25: 1.0}
    
    qf_15 = QuantumFloat(3)
    qf_15[:] = 4
    qf_16 = QuantumFloat(3)
    qf_16[:] = 3
    qf_15 += qf_16
    qf_15 -= 2
    assert qf_15.get_measurement() == {5.0: 1.0}
    


def test_ConditionEnvironment():
    
    from qrisp import QuantumChar, QuantumFloat, QuantumBool, QuantumVariable, h, x, cx, mcx, p, multi_measurement, ConditionEnvironment
    import numpy as np
    
    q_ch = QuantumChar()
    qf = QuantumFloat(3, signed = True)

    h(q_ch[0])

    with q_ch == "a":
        qf += 2
        
    with q_ch == "a" as cond_bool:
        qf += 2
        cond_bool.flip()
        qf -= 2
        p(np.pi, cond_bool)
    
    assert multi_measurement([q_ch,qf]) == {('a', 4): 0.5, ('b', -2): 0.5}
    

    def quantum_eq(qv_0, qv_1):

        res = QuantumBool(name = "res_bool")

        if qv_0.size != qv_1.size:
            raise Exception("Tried to evaluate equality condition for QuantumVariables of differing size")

        temp_qv = QuantumVariable(qv_0.size)

        cx(qv_0, temp_qv)
        cx(qv_1, temp_qv)
        x(temp_qv)

        mcx(temp_qv, res)

        return res

    #Create some sample arguments on which to evaluate the condition
    q_bool_0 = QuantumBool()
    q_bool_1 = QuantumBool()
    q_bool_2 = QuantumBool()

    h(q_bool_0)

    with ConditionEnvironment(cond_eval_function = quantum_eq, args = [q_bool_0, q_bool_1]):
        q_bool_2.flip()
        
    assert multi_measurement([q_bool_0, q_bool_1, q_bool_2]) == {(False, False, True): 0.5, (True, False, False): 0.5}
    

def test_InversionEnvironment():
    
    from qrisp import QuantumFloat, invert, QuantumBool, h, q_mult, multi_measurement
    
    qf = QuantumFloat(3)
    
    h(qf)
    
    mult_res = q_mult(qf, qf)
    
    q_bool = QuantumBool()
    
    with mult_res < 10:
        q_bool.flip()
    
    with invert():
        q_mult(qf, qf, target = mult_res)
    
    mult_res.delete(verify = True)
    
    assert multi_measurement([qf, q_bool]) == {(0, True): 0.125, (1, True): 0.125, (2, True): 0.125, (3, True): 0.125, (4, False): 0.125, (5, False): 0.125, (6, False): 0.125, (7, False): 0.125}




"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

import weakref


#This module contains the necessary tools to merge QuantumSessions

#Due to the interplay with QuantumEnvironments and the behavior
#that QuantumEnvironments automatically detect QuantumSessions that operated
#within them there is quite a bit of complexity involved here.

#The idea to merge two session qs_0 and qs_1 is to first add the appropriate 
#amount of qubits/clbits to qs_0, append all the instructions and 
#fix the other attributes Then we let all the attributes of qs_1 point 
#to the corresponding attributes of qs_0 such that any change on one quantum session
#also appears in the other quantum session

#Therefore two QuantumSessions are declared as identical (or merged) if their attributes
#(like .data, .qubits) point to the same lists as the other one. 
#The == operator for QuantumSessions checks if the .data attribute points to the same list

#If a merge is requested, the first step is to "engage" the QuantumSessions
#in the environment stack. By default QuantumEnvironments create their own QuantumSession
#and "engaging" here means that if merge is called on a QuantumSession within an environment,
#this QuantumSession will be merged with all the QuantumSessions of the environment stack.
def merge_sessions(qs_0, qs_1):
    
    #We now iterate through every QuantumEnvironment and insert it's data into the correct original_data attribute
    for qs in [qs_0, qs_1]:
        
        #We need to find the environment where the env_qs quantum session is not merged into qs.
        #This implies that the instructions of qs have been appended in this environment's parent.
        #Therefore all the data needs to go into the original_data attribute of this environment.
        for env in qs.env_stack:
            if not env.env_qs == qs:
                env.original_data.extend(qs.data)
                qs.data = []
                merge_sessions_inner(qs, env.env_qs)

    merge_sessions_inner(qs_0, qs_1)
    
    

def merge_sessions_inner(qs_0, qs_1, merge_env_stack_ = True):
    if qs_0 == qs_1:
        return
    
    from qrisp.quantum_network import QuantumNetworkSession
    
    
    if isinstance(qs_0, QuantumNetworkSession):
        
        if not qs_0.backend is qs_1.backend:
            if isinstance(qs_1, QuantumNetworkSession):
                raise Exception("Tried to merge two QuantumNetworkSessions containing differing clients")
            
            if not qs_1.backend is None:
                raise Exception("Tried to merge a QuantumNetworkSession with a QuantumSession with a non-trivial backend")
            
        qs_1.backend = qs_0.backend
        qs_1.inbox = qs_0.inbox
        qs_1.__class__ = qs_0.__class__
    
    if isinstance(qs_1, QuantumNetworkSession):
        
        if not qs_1.backend is qs_0.backend:
            
            if isinstance(qs_0, QuantumNetworkSession):
                raise Exception("Tried to merge two QuantumNetworkSessions containing differing clients")
            
            if not qs_0.backend is None:
                raise Exception("Tried to merge a QuantumNetworkSession with a QuantumSession with a non-trivial backend")
            
        qs_0.backend = qs_1.backend
        qs_0.inbox = qs_1.inbox
        qs_0.__class__ = qs_1.__class__
    
    
    
    #The problem we face here is the following:
    #If two sessions a, b are merged and b is merged with another session c
    #at a later point, the attributes of b are updated by that merge, 
    #but the attributes of a are not. To solve this problem, we declare one of the merge
    #participants the "parent" of the merge. The parent has an attribute called "shadow_sessions"
    #that contain the sessions that the parent has been merged with. 
    #The attribute updates are then performed on all shadow sessions.
    
    #To make sure we only operate on parent sessions, we check the parent attribute
    while hasattr(qs_0, "parent"):
        qs_0 = qs_0.parent
    
    
    while hasattr(qs_1, "parent"):
        qs_1 = qs_1.parent
    
    resolve_naming_collisions(qs_0, qs_1)
    
    intersecting_qubits = set([qb.identifier for qb in qs_0.qubits]).intersection([qb.identifier for qb in qs_1.qubits])
    
    if len(intersecting_qubits):
        raise Exception(f"Tried to merge two QuantumSessions containing identically named qubits {intersecting_qubits}")
    
    if len(qs_0.env_stack) < len(qs_1.env_stack):
        qs_0, qs_1 = qs_1, qs_0
        
        
    if merge_env_stack_:
        merge_env_stack(qs_0, qs_1)
    
        if qs_0 == qs_1:
            return
    
    #The qs_tracker list should only contain one Session instance per equivalence class.
    #Ie. If two sessions are merged only one of them should appear in the the qs_tracker
    i = 0
    while i < len(qs_0.qs_tracker):
        if qs_0.qs_tracker[i]() is None:
            qs_0.qs_tracker.pop(i)
        elif qs_0.qs_tracker[i]() == qs_1:
            qs_0.qs_tracker.pop(i)
        else:
            i+= 1

    #Each environment contains a list of QuantumSessions that when it was entered.
    #This list determines if a QuantumSession is engaged into an QuantumEnvironment, if merge
    #is called from inside the QuantumEnvironment.
    #We only want to have one of the two QuantumSessions that are merged in this list, which is  
    #resolved by the identify_sessions_in_environment
    for i in range(len(qs_1.env_stack)):
        identify_sessions_in_environment(qs_1.env_stack[-i-1], qs_0, qs_1)
    
    
    if not qs_0.backend is None and not qs_1.backend is None:
        if id(qs_0.backend) != id(qs_1.backend):
            raise Exception("Tried to merge QuantumSessions with differing, non-trivial default backends.")
    
    if qs_0.backend is None:
        qs_0.backend = qs_1.backend

    if set(qs_0.clbits).intersection(qs_1.clbits):
        raise Exception("Tried to merge sessions with common classical bits")
    
    for qb in qs_1.qubits: qb.qs = weakref.ref(qs_0)
    for cb in qs_1.clbits: cb.qs = weakref.ref(qs_0)
    
    qs_0.qubits.extend(qs_1.qubits)
    qs_0.clbits.extend(qs_1.clbits)
    
    qs_0.data.extend(qs_1.data)
    
    object.__setattr__(qs_1, "data", qs_0.data)    
    
    #Update qv_list
    while len(qs_1.qv_list):
        
        qv = qs_1.qv_list.pop()
        #Patch quantum session attribute
        qv.qs = qs_0
        
        qs_0.qv_list.append(qv)
    
    
    qs_0.deleted_qv_list.extend(qs_1.deleted_qv_list)
    
    reorder_quantum_variables(qs_0)
    
    #Add variables to the uncomputation stack
    qs_0.uncomp_stack.extend(qs_1.uncomp_stack)
    
    qs_0.shadow_sessions.extend(qs_1.shadow_sessions + [weakref.ref(qs_1)])
    
    i = 0
    while i < len(qs_0.shadow_sessions):
        qs = qs_0.shadow_sessions[i]()
        if qs is None:
            qs_0.shadow_sessions.pop(i)
            continue
        
        qs.parent = qs_0
        
        # Update all the attributes
        for var in qs_0.__dict__.keys():
        
            qs.__dict__[var] = qs_0.__dict__[var]
            
        i += 1
            
            

def resolve_naming_collisions(qs_0, qs_1):
    
    qs_names_0 = [qv.name for qv in qs_0.qv_list + qs_0.deleted_qv_list]
    qs_names_1 = [qv.name for qv in qs_1.qv_list + qs_1.deleted_qv_list]
    
    intersecting_qv_names = set(qs_names_0).intersection(qs_names_1)

    if len(intersecting_qv_names):
        
        for qv_name in intersecting_qv_names:
            
            
            qv_0_index = qs_names_0.index(qv_name)
            
            qv_0 = (qs_0.qv_list + qs_0.deleted_qv_list)[qv_0_index]
            
            qv_1_index = qs_names_1.index(qv_name)
            qv_1 = (qs_1.qv_list + qs_1.deleted_qv_list)[qv_1_index]
        
            if qv_1.user_given_name:
                
                if qv_0.user_given_name:
                    raise Exception("Tried to merge QuantumSession containing "
                                    f"identically named QuantumVariables {qv_1.name}")
                
                qv_0, qv_1 = qv_1, qv_0
                
            else: 
                if qv_0.creation_time > qv_1.creation_time:
                    if not qv_1.user_given_name:
                        qv_0, qv_1 = qv_1, qv_0
        
        
            proposed_new_name = qv_1.name + "_1"
            k = 1
            
            while proposed_new_name in qs_names_0 + qs_names_1:
                k += 1
                proposed_new_name = qv_1.name + "_" + str(k)
            
            qv_1.name = proposed_new_name
            
            
            for k in range(len(qv_1.reg)):
                qv_1.reg[k].identifier = qv_1.name + "." + str(k)


def merge_env_stack(qs_0, qs_1):
    
    #We need to find the environment where the env_qs quantum session is not merged into qs.
    #This implies that the instructions of this session have been appended in this environment's parent.
    #Therefore all the data needs to go into the original_data attribute of this environment.
    for env in qs_0.env_stack:
        
        if env.env_qs == qs_0 and not env.env_qs == qs_1:
            if env in qs_1.env_stack:
                env.original_data.extend(qs_1.data)
                qs_1.data = []
            
        if env.env_qs == qs_1 and not env.env_qs == qs_0:
            for env_ in qs_0.env_stack:
                if not env_.env_qs == qs_0:
                    
                    env_.original_data.extend(qs_0.data)
                    qs_0.data = []
                    merge_sessions_inner(qs_0, env_.env_qs, False)
                if env.env_qs == env_.env_qs:
                    break

def reorder_quantum_variables(qs):

    overall_qv_list = qs.qv_list + qs.deleted_qv_list
    
    qs.qv_list.sort(key = lambda x : x.creation_time)
    
    overall_qv_list.sort(key = lambda x : x.creation_time)
    
    sorted_qubit_list = []
    for qv in overall_qv_list:
        sorted_qubit_list += qv.reg
        
    sorted_qubit_list.extend(list(set(qs.qubits) - set(sorted_qubit_list)))
    
    qs.qubits = sorted_qubit_list


def identify_sessions_in_environment(env, qs_0, qs_1):
    
    i = 0
    while i < len(env.active_qs_list):
        if env.active_qs_list[i] == qs_1:
            env.active_qs_list.pop(i)
            continue
        i += 1
    


def multi_session_merge(input_list):
    
    session_list = input_list
    session_list.sort(key = lambda x : -len(x.data))
    
    if not len(session_list):
        from qrisp.core import QuantumSession
        return QuantumSession()
    
    
    for i in range(1, len(session_list)):
        merge_sessions(session_list[0], session_list[i])
    
    return session_list[0]

def recursive_qs_search(input):
        
    if isinstance(input, str):
        return []
    
    from qrisp.core import QuantumSession
    
    if hasattr(input, "__iter__"):
        iterable = True
        
    #This case would in principle allow for also searching objects
    #this however requires a considerable chunk of resources
    elif hasattr(input, "__dict__") and not isinstance(input, QuantumSession):
        iterable = False
        # input = input.__dict__
    else:
        iterable = False
    
    result = []
    if iterable:
        if isinstance(input, dict):
            for key in input.keys():
                result += recursive_qs_search(key)
                result += recursive_qs_search(input[key])
        else:
            for i in range(len(input)):
                result += recursive_qs_search(input[i])
    else:
        if isinstance(input, QuantumSession):
            result = [input]
        else:
            try:
                if isinstance(input(), QuantumSession):
                    result = [input()]
            except TypeError:
                pass
            
            try:
                if isinstance(input.qs(), QuantumSession):
                    result = [input.qs()]
            except AttributeError:
                pass
            
            

    return result

def recursive_qv_search(input):
        
    if isinstance(input, str):
        return []
    from qrisp.core import QuantumVariable
    
    if hasattr(input, "__iter__"):
        iterable = True
    elif hasattr(input, "__dict__") and not isinstance(input, QuantumVariable):
        iterable = False
        # input = input.__dict__
    else:
        iterable = False
        
    result = []
    if iterable:
        if isinstance(input, dict):
            for key in input.keys():
                result += recursive_qv_search(key)
                result += recursive_qv_search(input[key])
        else:
            for i in range(len(input)):
                result += recursive_qv_search(input[i])
    else:
        if isinstance(input, QuantumVariable):
            result = [input]

    return result


def merge(*args):
    session_list = recursive_qs_search(args)
    return multi_session_merge(session_list)



"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

# Created by ann81984 at 07.05.2022
import time

from qrisp.core import QuantumSession
from qrisp.arithmetic import QuantumFloat
from numpy.linalg import norm
from qiskit import execute, Aer
from qrisp.interface import convert_circuit

def test_unitary_calculation():
    n = 2

    qs = QuantumSession()

    x = QuantumFloat(n, 0, qs, signed = True)
    y = QuantumFloat(n, 0, qs, signed = False)

    s = x*y
    #qs = s.qs
    qc = qs.compile()
    
    start = time.time()
    test_unitary_1 = qc.get_unitary()
    end = time.time()

    print("Qrisp calculation time = %s" % (end - start))



    ###################
    qc.qubits = qc.qubits[::-1]
    qiskit_qc = convert_circuit(qc, target_api = "qiskit")

    backend = Aer.get_backend('unitary_simulator')

    start = time.time()
    job = execute(qiskit_qc, backend)
    result = job.result()
    test_unitary_2 = result.get_unitary(qiskit_qc).data
    end = time.time()

    print("Qiskit calculation time = %s" % (end - start))

    #get the unitary matrix from the result object
    print("Unitaries matching:", bool(norm(test_unitary_1 - test_unitary_2) < 1E-3))
    assert bool(norm(test_unitary_1 - test_unitary_2) < 1E-3)


"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""


import numpy as np

def auto_uncompute(*args, recompute = False):
    if len(args):
        return auto_uncompute_inner(args[0])
    else:
        
        if recompute:
            from qrisp import gate_wrap
            def auto_uncompute_helper(function):
                return gate_wrap(auto_uncompute_inner(function))
            return auto_uncompute_helper
        else:
            return auto_uncompute_inner(args[0])


#Decorator for auto uncomputed function
def auto_uncompute_inner(function):
    
    #Create auto uncomputed function
    def auto_uncomputed_function(*args, **kwargs):
        
        from qrisp.core import multi_session_merge, recursive_qs_search, recursive_qv_search, QuantumVariable
        
        #Acquire quantum session
        
        qs_list = recursive_qs_search([args, kwargs])
        multi_session_merge(qs_list)
        
        if not len(qs_list):
            return function(*args, **kwargs)
        
        qs = qs_list[0]
        
        #Determine quantum variables to uncompute
        
        initial_qvs = set([hash(qv()) for qv in QuantumVariable.live_qvs])
        
        #Execute function
        result = function(*args, **kwargs)
        
        result_vars = set([hash(qv) for qv in recursive_qv_search(result)])
        
        uncomp_vars = []
        
        for qv in qs.qv_list:
            
            if not hash(qv) in initial_qvs.union(result_vars):
                
                uncomp_vars.append(qv)
        
        uncompute(qs, uncomp_vars)
        
        return result
    
    auto_uncomputed_function.__name__ = function.__name__ + "_auto_uncomputed"
    
    #Return result
    return auto_uncomputed_function

    

verify = np.zeros(1)

def uncompute(qs, uncomp_vars, recompute = False):
    from qrisp import QuantumEnvironment
    
    qubits_to_uncompute = sum([qv.reg for qv in uncomp_vars], [])
    
    i = 0
    while i < len(qs.data):
        if isinstance(qs.data[i], QuantumEnvironment):
            env = qs.data.pop(i)
            env.compile()
            continue
        i += 1
    
    alloc_gates_remaining = list(qubits_to_uncompute)
    
    for i in range(len(qs.data)):
        
        instr = qs.data[i]
        if isinstance(instr, QuantumEnvironment):
            continue
        if instr.op.name == "qb_alloc" and instr.qubits[0] in alloc_gates_remaining:
            alloc_gates_remaining.remove(instr.qubits[0])
    
    if len(alloc_gates_remaining):
        
        incorrectly_allocated_qvs = []
        
        for qv in uncomp_vars:
            if set(alloc_gates_remaining).intersection(list(qv)):
                incorrectly_allocated_qvs.append(qv)
        
        incorrectly_allocated_qvs = [qv.name for qv in incorrectly_allocated_qvs]
        raise Exception(f"Could not uncompute QuantumVariables {incorrectly_allocated_qvs} because they were not created within this QuantumEnvironment")
            
        
    qc_to_uncompute = qs.copy()
    
    recompute_qubits = []
    
    for qb in qc_to_uncompute.qubits:
        if hasattr(qb, "recompute"):
            recompute_qubits.append(qb)
    
    from qrisp.uncomputation.unqomp import uncompute_qc
    
    uncomputed_qc = uncompute_qc(qc_to_uncompute, qubits_to_uncompute, recompute_qubits)
    
    
    from qrisp import QuantumCircuit
    
    qs.data = []
    QuantumCircuit.fast_append = True
    
    qs.data = uncomputed_qc.data
    
    for qv in uncomp_vars:
        
        # if recompute:
            # for qb in qv:
                # qb.recompute = True
        
        qv.delete(verify = bool(verify), recompute = recompute)
        
        #The uncomputation algorithm already added deallocation gates
        #Therefore we remove the deallocation gates
        for i in range(len(qv)):
            qv.qs.data.pop(-1)
            
    QuantumCircuit.fast_append = False


"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""


from qrisp.quantum_network.qn_server import QuantumNetworkServer

test = QuantumNetworkServer(socket_ip_address = "127.0.0.1", background = True)
test.start()


from qrisp.quantum_network.quantum_network_session import QuantumNetworkSession

charlie = QuantumNetworkSession(socket_ip = "127.0.0.1", port = 7070, name = "charlie")
bob = QuantumNetworkSession(socket_ip = "127.0.0.1", port = 7070, name = "bob")

from qrisp import QuantumVariable, x, h, QuantumFloat

print("Server creation done")
qf_1 = QuantumFloat(3, qs = bob)
qf_2 = QuantumFloat(3, qs = bob)

qf_1[:] = 2
qf_2[:] = 3

qf_3 = qf_2*qf_1

qf_3 += 3

print("Compilation done")

bob.send_qv(qf_3, "charlie")


print("Communication done")

inbox = charlie.update()

print(inbox[0])

from qrisp import QuantumChar, QuantumBool


q_ch = QuantumBool()
q_ch[:] = False

bob.send_qv(q_ch, "charlie")

inbox = charlie.update()

print(inbox[0])
